# 31Mg radioactive decay function for fitting β-NMR SLR spectra, see:
# R. M. L. McFadden et al. JPS Conf. Proc. 21, 011047 (2018).
# https://doi.org/10.7566/JPSCP.21.011047
# 
# General solutions to the rate equations computed with:
# Maxima, a Computer Algebra System
# http://maxima.sourceforge.net/
# 
# Original Author: Ryan M. L. McFadden
# Transcribed to python by: Derek Fujimoto
# June 2019

import numpy as np

# GLOBAL CONSTANTS ===========================================================

# half-lives (s)
T_12_31Mg = 0.235
T_12_31Al = 0.644
T_12_31Si = 2.632*3600  # seconds
T_12_31P = np.inf
T_12_30Al = 3.62
T_12_30Si = np.inf

# nuclear lifetimes (s)
ln2 = np.log(2)
tau_31Mg = T_12_31Mg / ln2
tau_31Al = T_12_31Al / ln2
tau_31Si = T_12_31Si / ln2
tau_31P =  T_12_31P  / ln2
tau_30Al = T_12_30Al / ln2
tau_30Si = T_12_30Si / ln2

# decay constants (1/s)
lambda_31Mg = 1/tau_31Mg
lambda_31Al = 1/tau_31Al
lambda_31Si = 1/tau_31Si
lambda_31P  = 1/tau_31P
lambda_30Al = 1/tau_30Al
lambda_30Si = 1/tau_30Si

# branching ratios
b_31Mg = 0.938
b_31Al = 0.984

# Q values (keV)
Q_31Mg_b = 11829.0
Q_31Mg_bn = 4671.0
Q_31Al_b = 7998.3
Q_31Al_bn = 1410.9
Q_31Si_b = 1491.50
Q_30Al_b = 8568.1

# effective Q values
Q_31Mg = b_31Mg * Q_31Mg_b + (1-b_31Mg) * Q_31Mg_bn
Q_31Al = b_31Al * Q_31Al_b + (1-b_31Al) * Q_31Al_bn
Q_31Si = Q_31Si_b
Q_30Al = Q_30Al_b

# find the maximum effective Q value
Q_max = max((Q_31Mg, Q_31Al, Q_31Si, Q_30Al))

# naive detection efficiencies based on effective Q values
e_31Mg = Q_31Mg / Q_max
e_31Al = Q_31Al / Q_max
e_31Si = Q_31Si / Q_max
e_31P = 0.0
e_30Al = Q_30Al / Q_max
e_30Si = 0.0

# FUNCTIONS ==================================================================
def s_31Mg(time, beam_rate, n_31Mg_0):
  return    ((n_31Mg_0 * lambda_31Mg - beam_rate) * np.exp(-time * lambda_31Mg)\
            ) / lambda_31Mg + beam_rate / lambda_31Mg

# number 31Mg atoms
def n_31Mg(time, beam_pulse, beam_rate):
    
    # get indexes for pre beam-off
    out = np.zeros(time.shape[0])
    idx = time<=beam_pulse
    
    # beam on 
    out[idx] = s_31Mg(time[idx], beam_rate, 0)
    
    # beam off
    delta = time[~idx] - beam_pulse
    n_31Mg_0 = s_31Mg(beam_pulse, beam_rate, 0)
    out[~idx] = s_31Mg(delta, 0, n_31Mg_0)
    
    # negative times
    out[time<0] = 0
    return out

# activity of 31Mg atoms
def a_31Mg(time, beam_pulse, beam_rate):
    return e_31Mg * lambda_31Mg * n_31Mg(time, beam_pulse, beam_rate)

def s_31Al(time, beam_rate, n_31Mg_0, n_31Al_0):
    
    nb_31Mg = n_31Mg_0 * b_31Mg
    rb_31Mg = beam_rate * b_31Mg
    
    term1 = -(nb_31Mg * lambda_31Mg - rb_31Mg) * np.exp(-time * lambda_31Mg) \
             / (lambda_31Mg - lambda_31Al)
               
    term2 = np.exp(-time * lambda_31Al) * \
            (((nb_31Mg + n_31Al_0) * lambda_31Al - rb_31Mg) * \
               lambda_31Mg - n_31Al_0 * lambda_31Al**2)
               
    return term1 + term2 / (lambda_31Al * lambda_31Mg - lambda_31Al**2) + \
           rb_31Mg / lambda_31Al

# number of 31Al atoms
def n_31Al(time, beam_pulse, beam_rate):
    
    # get indexes for pre beam-off
    out = np.zeros(time.shape[0])
    idx = time<=beam_pulse
    
    # beam on 
    out[idx] = s_31Al(time[idx], beam_rate, 0, 0)

    # beam off
    delta = time[~idx] - beam_pulse
    n_31Mg_0 = s_31Mg(beam_pulse, beam_rate, 0)
    n_31Al_0 = s_31Al(beam_pulse, beam_rate, 0, 0)
    out[~idx] = s_31Al(delta, 0, n_31Mg_0, n_31Al_0)

    # negative times
    out[time<0] = 0
    return out

# activity of 31Al atoms
def a_31Al(time, beam_pulse, beam_rate):
  return e_31Al * lambda_31Al * n_31Al(time, beam_pulse, beam_rate)

def s_31Si(time, beam_rate, n_31Mg_0, n_31Al_0, n_31Si_0):
    
    r_bAlbMg = beam_rate * b_31Al * b_31Mg
    lam_prod = lambda_31Al * lambda_31Mg
    delta_lam = lambda_31Mg - lambda_31Al
    
    prod1 = n_31Mg_0 * b_31Al * b_31Mg + n_31Al_0 * b_31Al
    prod2 = lam_prod * lambda_31Si
    
    
    
    return ((n_31Si_0 * lambda_31Si**3 +\
           ((-n_31Al_0 * b_31Al - n_31Si_0) * lambda_31Al -\
            n_31Si_0 * lambda_31Mg) * lambda_31Si**2 + (prod1 + n_31Si_0) * prod2-\
           r_bAlbMg * lam_prod) *\
          np.exp(-time * lambda_31Si)) /\
             (lambda_31Si**3 +\
              (-lambda_31Mg - lambda_31Al) * lambda_31Si**2 + prod2) -\
         ((n_31Mg_0 * b_31Al * b_31Mg * lam_prod -\
           r_bAlbMg * lambda_31Al) *\
          np.exp(-time * lambda_31Mg)) / (delta_lam * lambda_31Si -\
              lambda_31Mg**2 + lam_prod) +\
         (np.exp(-time * lambda_31Al) *\
          ((prod1 * lambda_31Al - r_bAlbMg) * lambda_31Mg -\
           n_31Al_0 * b_31Al * (lambda_31Al**2))) /\
             (delta_lam * lambda_31Si -\
              lam_prod + lambda_31Al**2) +r_bAlbMg / lambda_31Si

# number of 31Si atoms
def n_31Si(time, beam_pulse, beam_rate):
    
    # get indexes for pre beam-off
    out = np.zeros(time.shape[0])
    idx = time<=beam_pulse
    
    # beam on 
    out[idx] = s_31Si(time[idx], beam_rate, 0, 0, 0)

    # beam off
    delta = time[~idx] - beam_pulse
    n_31Mg_0 = s_31Mg(beam_pulse, beam_rate, 0)
    n_31Al_0 = s_31Al(beam_pulse, beam_rate, 0, 0)
    n_31Si_0 = s_31Si(beam_pulse, beam_rate, 0, 0, 0)
    out[~idx] = s_31Si(delta, 0, n_31Mg_0, n_31Al_0, n_31Si_0)
    
    # negative times
    out[time<0] = 0
    return out

# activity of 31Si atoms
def a_31Si(time, beam_pulse, beam_rate):
  return e_31Si * lambda_31Si * n_31Si(time, beam_pulse, beam_rate)

def s_31P(time, beam_rate, n_31Mg_0, n_31Al_0, n_31Si_0, n_31P_0):
    
    bb = b_31Al * b_31Mg
    rbb = beam_rate * bb
    bblamAl = bb * lambda_31Al
    r_bblamAl = beam_rate * bblamAl
    lamlam = lambda_31Al * lambda_31Mg
    
    return -(\
            (n_31Si_0 * lambda_31Si**3 +\
                ((-n_31Al_0 * b_31Al - n_31Si_0) * lambda_31Al -n_31Si_0 * lambda_31Mg) *\
                lambda_31Si**2 +\
                (n_31Mg_0 * bb + n_31Al_0 * b_31Al + n_31Si_0) *\
                lamlam * lambda_31Si -\
                r_bblamAl * lambda_31Mg) *\
            np.exp(-time * lambda_31Si)) /\
            (lambda_31Si**3 + (-lambda_31Mg - lambda_31Al) * lambda_31Si**2 +\
                lamlam * lambda_31Si) +\
            ((n_31Mg_0 * bblamAl * lambda_31Mg - r_bblamAl) *\
                np.exp(-time * lambda_31Mg) * lambda_31Si) /\
            ((lambda_31Mg**2 - lamlam) *\
            lambda_31Si - lambda_31Mg**3 + lambda_31Al * lambda_31Mg**2) -\
            (np.exp(-time * lambda_31Al) *\
                (((n_31Mg_0 * bb + n_31Al_0 * b_31Al) * lambda_31Al -rbb) *\
                    lambda_31Mg -n_31Al_0 * b_31Al * lambda_31Al**2) *\
                lambda_31Si) /\
            ((lamlam - lambda_31Al**2) *\
                  lambda_31Si - lambda_31Al**2 * lambda_31Mg + lambda_31Al**3) +\
            ((((n_31Mg_0 * bb + n_31Al_0 * b_31Al + n_31Si_0 + n_31P_0) *\
                lambda_31Al - rbb) * \
                lambda_31Mg - r_bblamAl) *\
                lambda_31Si - r_bblamAl * lambda_31Mg) /\
            (lamlam * lambda_31Si) + rbb * time

# number of 31P atoms
def n_31P(time, beam_pulse, beam_rate): 
    
    # get indexes for pre beam-off
    out = np.zeros(time.shape[0])
    idx = time<=beam_pulse
    
    # beam on 
    out[idx] = s_31P(time[idx], beam_rate, 0, 0, 0, 0)

    # beam off
    delta = time[~idx] - beam_pulse    
    n_31Mg_0 = s_31Mg(beam_pulse, beam_rate, 0)
    n_31Al_0 = s_31Al(beam_pulse, beam_rate, 0, 0)
    n_31Si_0 = s_31Si(beam_pulse, beam_rate, 0, 0, 0)
    n_31P_0 = s_31P(time[~idx], beam_rate, 0, 0, 0, 0)
    out[~idx] = s_31P(delta, 0, n_31Mg_0, n_31Al_0, n_31Si_0, n_31P_0)
    
    # negative times
    out[time<0] = 0
    return out

def a_31P(time, beam_pulse, beam_rate): 
    return e_31P * lambda_31P * n_31P(time, beam_pulse, beam_rate)

def s_30Al( time,  beam_rate,  n_31Mg_0,  n_30Al_0): 
    return  (((n_31Mg_0 * b_31Mg - n_31Mg_0) * \
            lambda_31Mg - beam_rate * b_31Mg +beam_rate) *\
            np.exp(-time * lambda_31Mg)) /\
            (lambda_31Mg - lambda_30Al) -\
            (np.exp(-time * lambda_30Al) *\
            (((n_31Mg_0 * b_31Mg - n_31Mg_0 - n_30Al_0) * lambda_30Al -\
            beam_rate * b_31Mg + beam_rate) *\
            lambda_31Mg + n_30Al_0 * lambda_30Al**2)) /\
            (lambda_30Al * lambda_31Mg - lambda_30Al**2) -\
            (beam_rate * b_31Mg - beam_rate) / lambda_30Al

# number of 30Al atoms
def n_30Al(time, beam_pulse, beam_rate):
    
    # get indexes for pre beam-off
    out = np.zeros(time.shape[0])
    idx = time<=beam_pulse
    
    # beam on 
    out[idx] = s_30Al(time[idx], beam_rate, 0, 0)
  
    # beam off
    delta = time[~idx] - beam_pulse    
    n_31Mg_0 = s_31Mg(beam_pulse, beam_rate, 0)
    n_30Al_0 = s_30Al(beam_pulse, beam_rate, 0, 0)
    out[~idx] = s_30Al(delta, 0, n_31Mg_0, n_30Al_0)
  
    # negative times
    out[time<0] = 0
    return out
    
def a_30Al(time, beam_pulse, beam_rate): 
  return e_30Al * lambda_30Al * n_30Al(time, beam_pulse, beam_rate)

def s_30Si(time, beam_rate, n_31Mg_0, n_31Al_0, n_30Al_0, n_30Si_0):
    return -((((n_31Mg_0 * b_31Al - n_31Mg_0) * b_31Mg * lambda_31Al +\
            (n_31Mg_0 * b_31Mg - n_31Mg_0) * lambda_30Al) *\
            lambda_31Mg**2 +\
            (((n_31Mg_0 - n_31Mg_0 * b_31Al * b_31Mg) * lambda_30Al +\
            (beam_rate - beam_rate * b_31Al) * b_31Mg) *\
            lambda_31Al +\
            (beam_rate - beam_rate * b_31Mg) * lambda_30Al) *\
            lambda_31Mg +\
            (beam_rate * b_31Al * b_31Mg - beam_rate) * lambda_30Al *\
            lambda_31Al) *\
            np.exp(-time * lambda_31Mg)) /\
            (lambda_31Mg**3 +\
            (-lambda_31Al - lambda_30Al) * lambda_31Mg**2 +\
            lambda_30Al * lambda_31Al * lambda_31Mg) -\
            ((((n_31Mg_0 * b_31Al * b_31Mg + n_31Al_0 * b_31Al - n_31Mg_0 -\
            n_31Al_0 - n_30Si_0 - n_30Al_0) *\
            lambda_30Al -\
            beam_rate * b_31Mg + beam_rate) *\
            lambda_31Al +\
            (beam_rate - beam_rate * b_31Al) * b_31Mg * lambda_30Al) *\
            lambda_31Mg +\
            (beam_rate - beam_rate * b_31Al * b_31Mg) * lambda_30Al *\
            lambda_31Al) /\
            (lambda_30Al * lambda_31Al * lambda_31Mg) +\
            (np.exp(-time * lambda_31Al) *\
            ((((n_31Mg_0 * b_31Al - n_31Mg_0) * b_31Mg + n_31Al_0 * b_31Al -\
            n_31Al_0) *\
            lambda_31Al +\
            (beam_rate - beam_rate * b_31Al) * b_31Mg) *\
            lambda_31Mg +\
            (n_31Al_0 - n_31Al_0 * b_31Al) * lambda_31Al**2)) /\
            (lambda_31Al * lambda_31Mg - lambda_31Al**2) +\
            (np.exp(-time * lambda_30Al) *\
            (((n_31Mg_0 * b_31Mg - n_31Mg_0 - n_30Al_0) * lambda_30Al -\
            beam_rate * b_31Mg + beam_rate) *\
            lambda_31Mg +\
            n_30Al_0 * lambda_30Al**2)) /\
            (lambda_30Al * lambda_31Mg - lambda_30Al**2) - \
            beam_rate * b_31Al * b_31Mg * time + beam_rate * time

# number of 30Si atoms
def n_30Si(time, beam_pulse, beam_rate):
    
    # get indexes for pre beam-off
    out = np.zeros(time.shape[0])
    idx = time<=beam_pulse
    
    # beam on 
    out[idx] = s_30Si(time[idx], beam_rate, 0, 0, 0, 0)

    # beam off
    delta = time[~idx] - beam_pulse    
    n_31Mg_0 = s_31Mg(beam_pulse, beam_rate, 0)
    n_31Al_0 = s_31Al(beam_pulse, beam_rate, 0, 0)
    n_30Al_0 = s_30Al(beam_pulse, beam_rate, 0, 0)
    n_30Si_0 = s_30Si(beam_pulse, beam_rate, 0, 0, 0, 0)
    out[~idx] = s_30Si(delta, 0, n_31Mg_0, n_31Al_0, n_30Al_0, n_30Si_0)

    # negative times
    out[time<0] = 0
    return out
    
def a_30Si(time, beam_pulse, beam_rate): 
    return e_30Si * lambda_30Si * n_30Si(time, beam_pulse, beam_rate)

# total atoms
def n_total(time, beam_pulse, beam_rate=1e6):
    return  n_31Mg(time, beam_pulse, beam_rate) + \
            n_31Al(time, beam_pulse, beam_rate) + \
            n_31Si(time, beam_pulse, beam_rate) + \
            n_31P (time, beam_pulse, beam_rate) + \
            n_30Al(time, beam_pulse, beam_rate) + \
            n_30Si(time, beam_pulse, beam_rate);

# fractions of total atoms
def fn_31Mg(time, beam_pulse, beam_rate=1e6): 
    return  n_31Mg(time, beam_pulse, beam_rate) / \
            n_total(time, beam_pulse, beam_rate)

def fn_31Al(time, beam_pulse, beam_rate=1e6):
    return n_31Al(time, beam_pulse, beam_rate) / \
            n_total(time, beam_pulse, beam_rate)

def fn_31Si(time, beam_pulse, beam_rate=1e6):
  return n_31Si(time, beam_pulse, beam_rate) / \
         n_total(time, beam_pulse, beam_rate)

def fn_31P(time, beam_pulse, beam_rate=1e6):
  return n_31P(time, beam_pulse, beam_rate) / \
         n_total(time, beam_pulse, beam_rate)

def fn_30Al(time, beam_pulse, beam_rate=1e6):
  return n_30Al(time, beam_pulse, beam_rate) / \
         n_total(time, beam_pulse, beam_rate)

def fn_30Si(time, beam_pulse, beam_rate=1e6):
  return n_30Si(time, beam_pulse, beam_rate) / \
         n_total(time, beam_pulse, beam_rate)

# total activity
def a_total(time, beam_pulse, beam_rate=1e6):
  return a_31Mg(time, beam_pulse, beam_rate) + \
         a_31Al(time, beam_pulse, beam_rate) + \
         a_31Si(time, beam_pulse, beam_rate) + \
         a_31P (time, beam_pulse, beam_rate) + \
         a_30Al(time, beam_pulse, beam_rate) + \
         a_30Si(time, beam_pulse, beam_rate)

# fractional activities
def fa_31Mg(time, beam_pulse, beam_rate=1e6):
  return a_31Mg(time, beam_pulse, beam_rate) / \
         a_total(time, beam_pulse, beam_rate)

def fa_31Al(time, beam_pulse, beam_rate=1e6):
  return a_31Al(time, beam_pulse, beam_rate) / \
         a_total(time, beam_pulse, beam_rate)

def fa_31Si(time, beam_pulse, beam_rate=1e6):
  return a_31Si(time, beam_pulse, beam_rate) / \
         a_total(time, beam_pulse, beam_rate)

def fa_31P(time, beam_pulse, beam_rate=1e6):
  return a_31P(time, beam_pulse, beam_rate) / \
         a_total(time, beam_pulse, beam_rate)

def fa_30Al(time, beam_pulse, beam_rate=1e6):
  return a_30Al(time, beam_pulse, beam_rate) / \
         a_total(time, beam_pulse, beam_rate)

def fa_30Si(time, beam_pulse, beam_rate=1e6):
  return a_30Si(time, beam_pulse, beam_rate) / \
         a_total(time, beam_pulse, beam_rate)



# Drawing style window
# Derek Fujimoto
# July 2018

from tkinter import *
from tkinter import ttk
from bfit import logger_name
import webbrowser
import logging

# ========================================================================== #
class popup_redraw_period(object):
    """
        Popup window for setting redraw period. 
    """

    # ====================================================================== #
    def __init__(self, parent):
        self.parent = parent
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # make a new window
        self.win = Toplevel(parent.mainframe)
        self.win.title('Set Redraw Period')
        frame = ttk.Frame(self.win, relief='sunken', pad=5)
        topframe = ttk.Frame(frame, pad=5)

        # icon
        self.parent.set_icon(self.win)

        # Key bindings
        self.win.bind('<Return>', self.set)             
        self.win.bind('<KP_Enter>', self.set)

        # make objects: text entry
        l1 = ttk.Label(topframe, text='Redraw update period:', pad=5, justify=LEFT)
        self.text = IntVar()
        self.text.set(parent.update_period)
        entry = ttk.Entry(topframe, textvariable=self.text, width=10, justify=RIGHT)
        l2 = ttk.Label(topframe, text='s', pad=5, justify=LEFT)
        
        # make objects: buttons
        set_button = ttk.Button(frame, text='Set', command=self.set)
        close_button = ttk.Button(frame, text='Cancel', command=self.cancel)
        
        # grid
        l1.grid(column=0, row=0)
        entry.grid(column=1, row=0)
        l2.grid(column=2, row=0)
        topframe.grid(column=0, row=0, columnspan=2, pady=10)
        set_button.grid(column=0, row=1)
        close_button.grid(column=1, row=1)
            
        # grid frame
        frame.grid(column=0, row=0)
        self.logger.debug('Initialization success. Starting mainloop.')
    
    # ====================================================================== #
    def set(self, *args):
        """Set entered values"""        
        self.parent.update_period = self.text.get()
        self.logger.info('Set redraw period to %s s', self.parent.update_period)
        self.win.destroy()
        
    # ====================================================================== #
    def cancel(self):
        self.win.destroy()



# fit_files tab for bfit
# Derek Fujimoto
# Dec 2017

from tkinter import *
from tkinter import ttk, messagebox, filedialog
from functools import partial
from bdata import bdata, bmerged
from bfit import logger_name, __version__
from scipy.special import gamma, polygamma
from pandas.plotting import register_matplotlib_converters

from bfit.gui.calculator_nqr_B0 import current2field
from bfit.gui.popup_show_param import popup_show_param
from bfit.gui.popup_param import popup_param
from bfit.gui.popup_fit_results import popup_fit_results
from bfit.gui.popup_fit_constraints import popup_fit_constraints
from bfit.gui.popup_add_param import popup_add_param
from bfit.fitting.decay_31mg import fa_31Mg
from bfit.backend.entry_color_set import on_focusout, on_entry_click
from bfit.backend.raise_window import raise_window

import numpy as np
import pandas as pd
import bdata as bd
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import bfit.backend.colors as colors

import datetime, os, traceback, warnings, logging, yaml, textwrap

register_matplotlib_converters()

# =========================================================================== #
# =========================================================================== #
class fit_files(object):
    """
        Data fields:
            annotation:     stringvar: name of quantity for annotating parameters 
            asym_type:      asymmetry calculation type
            canvas_frame_id:id number of frame in canvas
            chi_threshold:  if chi > thres, set color to red
            draw_components:list of titles for labels, options to export, draw.
            entry_asym_type:combobox for asym calculations
            fit_canvas:     canvas object allowing for scrolling 
            par_label_entry:draw parameter label entry box
            pop_fitconstr:  object for fitting with constrained functions
            fit_data_tab:   containing frame (for destruction)
            fit_function_title: StringVar, title of fit function to use
            fit_function_title_box: combobox for fit function names
            fit_input:      fitting input values = (fn_name, ncomp, data_list)
            fit_lines:      Dict storing fitline objects
            fit_lines_old: dictionary of previously used fitline objects, keyed by run
            fit_routine_label: label for fit routine
            fitter:         fitting object from self.bfit.routine_mod
            gchi_label:     Label for global chisquared    
            mode:           what type of run is this. 
            
            n_component:    number of fitting components (IntVar)
            n_component_box:Spinbox for number of fitting components
            par_label       StringVar, label for plotting parameter set
            plt:            self.bfit.plt
            
            pop_addpar:     popup for ading parameters which are combinations of others
            pop_fitres:     modelling popup, for continuity between button presses
            pop_fitcontr:   popup for fitting constrained values
            
            probe_label:    Label for probe species
            runframe:       frame for displaying fit results and inputs
            runmode_label:  display run mode 
            set_as_group:   BooleanVar() if true, set fit parfor whole group
            set_prior_p0:   BooleanVar() if true, set P0 of newly added runs to 
                            P0 of fit with largest run number
            share_var:      BooleanVar() holds share checkbox for all fitlines
            use_rebin:      BoolVar() for rebinning on fitting
            xaxis:          StringVar() for parameter to draw on x axis
            yaxis:          StringVar() for parameter to draw on y axis
            xaxis_combobox: box for choosing x axis draw parameter
            yaxis_combobox: box for choosing y axis draw parameter
            
            xlo, hi:         StringVar, fit range limits on x axis
    """ 
    
    default_fit_functions = {
            '20':('Exp', 'Str Exp'), 
            '2h':('Exp', 'Str Exp'), 
            '1f':('Lorentzian', 'Gaussian'), 
            '1w':('Lorentzian', 'Gaussian'), 
            '1n':('Lorentzian', 'Gaussian')}
    mode = ""
    chi_threshold = 1.5 # threshold for red highlight on bad fits 
    n_fitx_pts = 500    # number of points to draw in fitted curves
    
    # ======================================================================= #
    def __init__(self, fit_data_tab, bfit):
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.debug('Initializing')
        
        # initialize
        self.bfit = bfit
        self.fit_output = {}
        self.share_var = {}
        self.fitter = self.bfit.routine_mod.fitter(keyfn = bfit.get_run_key, 
                                                   probe_species = bfit.probe_species.get())
        self.draw_components = list(bfit.draw_components)
        self.fit_data_tab = fit_data_tab
        self.plt = self.bfit.plt
        
        # additional button bindings
        self.bfit.root.bind('<Control-Key-u>', self.update_param)
        
        # make top level frames
        mid_fit_frame = ttk.Labelframe(fit_data_tab, 
                                       text='Set Initial Parameters', pad=5)
                    
        mid_fit_frame.grid(column=0, row=1, rowspan=6, sticky=(S, W, E, N), padx=5, pady=5)
        
        fit_data_tab.grid_columnconfigure(0, weight=1)   # fitting space
        fit_data_tab.grid_rowconfigure(6, weight=1)      # push bottom window in right frame to top
        mid_fit_frame.grid_columnconfigure(0, weight=1)
        mid_fit_frame.grid_rowconfigure(0, weight=1)
        
        # TOP FRAME -----------------------------------------------------------
        
        # fit function select 
        fn_select_frame = ttk.Labelframe(fit_data_tab, text='Fit Function')
        self.fit_function_title = StringVar()
        self.fit_function_title.set("")
        self.fit_function_title_box = ttk.Combobox(fn_select_frame, 
                textvariable=self.fit_function_title, state='readonly')
        self.fit_function_title_box.bind('<<ComboboxSelected>>', 
            lambda x :self.populate_param(force_modify=True))
        
        # number of components in fit spinbox
        self.n_component = IntVar()
        self.n_component.set(1)
        self.n_component_box = Spinbox(fn_select_frame, from_=1, to=20, 
                textvariable=self.n_component, width=5, 
                command=lambda:self.populate_param(force_modify=True))
        
        # fit and other buttons
        fit_button = ttk.Button(fn_select_frame, text='        Fit        ', command=self.do_fit, \
                                pad=1)
        constraint_button = ttk.Button(fn_select_frame, text='Constrained Fit', 
                                       command=self.do_fit_constraints, pad=1)
        set_param_button = ttk.Button(fn_select_frame, text='   Set Result as P0   ', 
                        command=self.do_set_result_as_initial, pad=1)                        
        reset_param_button = ttk.Button(fn_select_frame, text='     Reset P0     ', 
                        command=self.do_reset_initial, pad=1)
            
        # GRIDDING
            
        # top frame gridding
        fn_select_frame.grid(column=0, row=0, sticky=(W, E, N), padx=5, pady=5)
        
        c = 0
        self.fit_function_title_box.grid(column=c, row=0, padx=5); c+=1
        ttk.Label(fn_select_frame, text="Number of Terms:").grid(column=c, 
                  row=0, padx=5, pady=5, sticky=W); c+=1
        self.n_component_box.grid(column=c, row=0, padx=5, pady=5, sticky=W); c+=1
        fit_button.grid(column=c, row=0, padx=5, pady=1, sticky=W); c+=1
        constraint_button.grid(column=c, row=0, padx=5, pady=1, sticky=(W, E)); c+=1
        set_param_button.grid(column=c, row=0, padx=5, pady=1, sticky=W); c+=1
        reset_param_button.grid(column=c, row=0, padx=5, pady=1, sticky=W); c+=1
        
        # MID FRAME -----------------------------------------------------------
        
        # Scrolling frame to hold fitlines
        yscrollbar = ttk.Scrollbar(mid_fit_frame, orient=VERTICAL)         
        self.fit_canvas = Canvas(mid_fit_frame, bd=0,                # make a canvas for scrolling
                yscrollcommand=yscrollbar.set,                      # scroll command receive
                scrollregion=(0, 0, 5000, 5000), confine=True)       # default size
        yscrollbar.config(command=self.fit_canvas.yview)            # scroll command send
        self.runframe = ttk.Frame(self.fit_canvas, pad=5)           # holds 
        
        self.canvas_frame_id = self.fit_canvas.create_window((0, 0),    # make window which can scroll
                window=self.runframe, 
                anchor='nw')
        self.runframe.bind("<Configure>", self.config_canvas) # bind resize to alter scrollable region
        self.fit_canvas.bind("<Configure>", self.config_runframe) # bind resize to change size of contained frame
        
        # gridding
        self.fit_canvas.grid(column=0, row=0, sticky=(E, W, S, N))
        yscrollbar.grid(column=1, row=0, sticky=(W, S, N))
        
        self.runframe.grid_columnconfigure(0, weight=1) 
        self.fit_canvas.grid_columnconfigure(0, weight=1) 
        self.fit_canvas.grid_rowconfigure(0, weight=1)
        
        self.runframe.bind("<Configure>", self.config_canvas) # bind resize to alter scrollable region
        self.fit_canvas.bind("<Configure>", self.config_runframe) # bind resize to change size of contained frame
        
        # RIGHT FRAME ---------------------------------------------------------
        
        # run mode 
        fit_runmode_label_frame = ttk.Labelframe(fit_data_tab, pad=(10, 5, 10, 5), 
                text='Run Mode', )
        self.fit_runmode_label = ttk.Label(fit_runmode_label_frame, text="", justify=CENTER)
        
        # fitting routine
        fit_routine_label_frame = ttk.Labelframe(fit_data_tab, pad=(10, 5, 10, 5), 
                text='Minimizer', )
        self.fit_routine_label = ttk.Label(fit_routine_label_frame, text="", 
                                           justify=CENTER)
        
        # probe species
        probe_label_frame = ttk.Labelframe(fit_data_tab, pad=(10, 5, 10, 5), 
                text='Probe', )
        self.probe_label = ttk.Label(probe_label_frame, 
                                     text=self.bfit.probe_species.get(), 
                                     justify=CENTER)
        
        # global chisquared
        gchi_label_frame = ttk.Labelframe(fit_data_tab, pad=(10, 5, 10, 5), 
                text='Global ChiSq', )
        self.gchi_label = ttk.Label(gchi_label_frame, text='', justify=CENTER)
                     
        # asymmetry calculation
        asym_label_frame = ttk.Labelframe(fit_data_tab, pad=(60, 5, 5, 5), 
                text='Asymmetry Calculation', )
        self.asym_type = StringVar()
        self.asym_type.set('')
        self.entry_asym_type = ttk.Combobox(asym_label_frame, \
                textvariable=self.asym_type, state='readonly', \
                width=20)
        self.entry_asym_type['values'] = ()
        
        # other settings
        other_settings_label_frame = ttk.Labelframe(fit_data_tab, pad=(10, 5, 10, 5), 
                text='Switches', )
                
        # set as group checkbox
        self.set_as_group = BooleanVar()
        set_group_check = ttk.Checkbutton(other_settings_label_frame, 
                text='Modify for all', \
                variable=self.set_as_group, onvalue=True, offvalue=False)
        self.set_as_group.set(False)

        # rebin checkbox
        self.use_rebin = BooleanVar()
        set_use_rebin = ttk.Checkbutton(other_settings_label_frame, 
                text='Rebin data (set in fetch)', \
                variable=self.use_rebin, onvalue=True, offvalue=False)
        self.use_rebin.set(False)
        
        # set P0 as prior checkbox
        self.set_prior_p0 = BooleanVar()
        set_prior_p0 = ttk.Checkbutton(other_settings_label_frame, 
                text='Set P0 of new run to prior result', \
                variable=self.set_prior_p0, onvalue=True, offvalue=False)
        self.set_prior_p0.set(False)
        
        # specify x axis --------------------
        xspecify_frame = ttk.Labelframe(fit_data_tab, 
            text='Restrict x limits', pad=5)
        
        self.xlo = StringVar()
        self.xhi = StringVar()
        self.xlo.set('-inf')
        self.xhi.set('inf')
        
        entry_xspecify_lo = ttk.Entry(xspecify_frame, textvariable=self.xlo, width=10)
        entry_xspecify_hi = ttk.Entry(xspecify_frame, textvariable=self.xhi, width=10)
        label_xspecify = ttk.Label(xspecify_frame, text=" < x < ")
        
        # fit results -----------------------
        results_frame = ttk.Labelframe(fit_data_tab, 
            text='Fit Results and Run Conditions', pad=5)     # draw fit results
        
        # draw and export buttons
        button_frame = Frame(results_frame)
        draw_button = ttk.Button(button_frame, text='Draw', command=self.draw_param)
        update_button = ttk.Button(button_frame, text='Update', command=self.update_param)
        export_button = ttk.Button(button_frame, text='Export', command=self.export)
        show_button = ttk.Button(button_frame, text='Compare', command=self.show_all_results)
        model_fit_button = ttk.Button(button_frame, text='Fit a\nModel', 
                                      command=self.do_fit_model)
        
        
        # menus for x and y values
        x_button = ttk.Button(results_frame, text="x axis:", command=self.do_add_param, pad=0)
        y_button = ttk.Button(results_frame, text="y axis:", command=self.do_add_param, pad=0)
        ann_button = ttk.Button(results_frame, text=" Annotation:", command=self.do_add_param, pad=0)
        label_label = ttk.Label(results_frame, text="Label:")
        
        self.xaxis = StringVar()
        self.yaxis = StringVar()
        self.annotation = StringVar()
        self.par_label = StringVar()
        
        self.xaxis.set('')
        self.yaxis.set('')
        self.annotation.set('')
        self.par_label.set('')
        
        self.xaxis_combobox = ttk.Combobox(results_frame, textvariable=self.xaxis, 
                                      state='readonly', width=19)
        self.yaxis_combobox = ttk.Combobox(results_frame, textvariable=self.yaxis, 
                                      state='readonly', width=19)
        self.annotation_combobox = ttk.Combobox(results_frame, 
                                      textvariable=self.annotation, 
                                      state='readonly', width=19)
        self.par_label_entry = ttk.Entry(results_frame, 
                                    textvariable=self.par_label, width=21)
        
        # gridding
        button_frame.grid(column=0, row=0, columnspan=2)
        draw_button.grid(column=0, row=0, padx=5, pady=5)
        update_button.grid(column=0, row=1, padx=5, pady=5)
        show_button.grid(column=1, row=0, padx=5, pady=5)
        export_button.grid(column=1, row=1, padx=5, pady=5)
        model_fit_button.grid(column=2, row=0, rowspan=2, pady=5, sticky=(N, S))
        
        x_button.grid(column=0, row=1, sticky=(E, W), padx=5)
        y_button.grid(column=0, row=2, sticky=(E, W), padx=5)
        ann_button.grid(column=0, row=3, sticky=(E, W), padx=5)
        label_label.grid(column=0, row=4, sticky=(E, W), padx=10)
        
        self.xaxis_combobox.grid(column=1, row=1, pady=5)
        self.yaxis_combobox.grid(column=1, row=2, pady=5)
        self.annotation_combobox.grid(column=1, row=3, pady=5)
        self.par_label_entry.grid(column=1, row=4, pady=5)
        
        # save/load state -----------------------
        state_frame = ttk.Labelframe(fit_data_tab, text='Program State', pad=5)
        state_save_button = ttk.Button(state_frame, text='Save', command=self.save_state)
        state_load_button = ttk.Button(state_frame, text='Load', command=self.load_state)
       
        state_save_button.grid(column=0, row=0, padx=5, pady=5)
        state_load_button.grid(column=1, row=0, padx=5, pady=5)
        
        # gridding
        fit_runmode_label_frame.grid(column=1, row=0, pady=5, padx=2, sticky=(N, E, W))
        self.fit_runmode_label.grid(column=0, row=0, sticky=(E, W))
        
        fit_routine_label_frame.grid(column=2, row=0, pady=5, padx=2, sticky=(N, E, W))
        self.fit_routine_label.grid(column=0, row=0, sticky=(E, W))
        
        probe_label_frame.grid(column=1, row=1, columnspan=1, sticky=(E, W, N), pady=2, padx=2)
        self.probe_label.grid(column=0, row=0)
        
        gchi_label_frame.grid(column=2, row=1, columnspan=1, sticky=(E, W, N), pady=2, padx=2)
        self.gchi_label.grid(column=0, row=0)
        
        asym_label_frame.grid(column=1, row=2, columnspan=2, sticky=(E, W, N), pady=2, padx=2)
        self.entry_asym_type.grid(column=0, row=0)
        
        other_settings_label_frame.grid(column=1, row=3, columnspan=2, sticky=(E, W, N), pady=2, padx=2)
        set_group_check.grid(column=0, row=0, padx=5, pady=1, sticky=W)
        set_use_rebin.grid(column=0, row=1, padx=5, pady=1, sticky=W)
        set_prior_p0.grid(column=0, row=2, padx=5, pady=1, sticky=W)
        
        entry_xspecify_lo.grid(column=0, row=0)
        label_xspecify.grid(column=1, row=0)
        entry_xspecify_hi.grid(column=2, row=0)
        
        xspecify_frame.grid(column=1, row=4, columnspan=2, sticky=(E, W, N), pady=2, padx=2)
        results_frame.grid(column=1, row=5, columnspan=2, sticky=(E, W, N), pady=2, padx=2)
        state_frame.grid(column=1, row=6, columnspan=2, sticky=(E, W, N), pady=2, padx=2)
        
        # resizing
        
        # fn select
        fn_select_frame.grid_columnconfigure(1, weight=1)    # Nterms label
        fn_select_frame.grid_columnconfigure(4, weight=100)    # constraints
        fn_select_frame.grid_columnconfigure(5, weight=1)  # set results as p0
        fn_select_frame.grid_columnconfigure(6, weight=1)  # reset p0
        
        # fitting frame
        self.fit_canvas.grid_columnconfigure(0, weight=1)    # fetch frame 
        self.fit_canvas.grid_rowconfigure(0, weight=1)
        
        # right frame
        for i in range(2):
            results_frame.grid_columnconfigure(i, weight=0)
        
        # store lines for fitting
        self.fit_lines = {}
        self.fit_lines_old = {}
        
    # ======================================================================= #
    def __del__(self):
        
        if hasattr(self, 'fit_lines'):       del self.fit_lines
        if hasattr(self, 'fit_lines_old'):   del self.fit_lines_old
        if hasattr(self, 'fitter'):          del self.fitter
        
        # kill buttons and frame
        try:
            for child in self.fetch_data_tab.winfo_children():
                child.destroy()
            self.fetch_data_tab.destroy()
        except Exception:
            pass
            
     # ======================================================================= #
    def _annotate(self, id, x, y, ptlabels, color='k', unique=True):
        """Add annotation"""
        
        # base case
        if ptlabels is None: return
        
        # do annotation
        for label, xcoord, ycoord in zip(ptlabels, x, y):        
            if type(label) != type(None):
                self.plt.annotate('param', id, label, 
                             xy=(xcoord, ycoord), 
                             xytext=(-3, 20), 
                             textcoords='offset points', 
                             ha='right', 
                             va='bottom', 
                             bbox=dict(boxstyle='round, pad=0.1', 
                                       fc=color, 
                                       alpha=0.1), 
                             arrowprops=dict(arrowstyle = '->', 
                                             connectionstyle='arc3, rad=0'), 
                             fontsize='xx-small', 
                             unique=unique
                            )    
   
    # ======================================================================= #
    def _make_shared_var_dict(self):
        """Make the dictionary to make sure all shared checkboxes are synched"""
        
        # get parameter list
        try:
            parlst = [p for p in self.fitter.gen_param_names(
                                                self.fit_function_title.get(), 
                                                self.n_component.get())]
                                                
        # no paramteters: empty out the variable list
        except KeyError:
            share_var = {}
        
        # make new shared list
        else:
            # re-initialize
            share_var = {p:BooleanVar() for p in parlst}    
            
            # set to old values if they exist
            for p in parlst:
                if p in self.share_var.keys():
                    share_var[p].set(self.share_var[p].get())
                    
        # save to object
        self.share_var = share_var
            
    # ======================================================================= #
    def canvas_scroll(self, event):
        """Scroll canvas with files selected."""
        if event.num == 4:
            self.fit_canvas.yview_scroll(-1, "units")
        elif event.num == 5:
            self.fit_canvas.yview_scroll(1, "units")  
    
    # ======================================================================= #
    def config_canvas(self, event):
        """Alter scrollable region based on canvas bounding box size. 
        (changes scrollbar properties)"""
        self.fit_canvas.configure(scrollregion=self.fit_canvas.bbox("all"))
    
    # ======================================================================= #
    def config_runframe(self, event):
        """Alter size of contained frame in canvas. Allows for inside window to 
        be resized with mouse drag""" 
        self.fit_canvas.itemconfig(self.canvas_frame_id, width=event.width)
    
    # ======================================================================= #
    def populate(self, *args):
        """
            Make tabs for setting fit input parameters. 
        """
        
        # get data
        dl = self.bfit.fetch_files.data_lines
        keylist = [k for k in dl.keys() if dl[k].check_state.get()]
        keylist.sort()
        self.logger.debug('Populating data for %s', keylist)
        
        # get run mode by looking at one of the data dictionary keys
        for key_zero in self.bfit.data.keys(): break
        
        # create fit function combobox options
        try:               
            if self.mode != self.bfit.data[key_zero].mode:
                
                # set run mode 
                self.mode = self.bfit.data[key_zero].mode 
                self.fit_runmode_label['text'] = \
                        self.bfit.fetch_files.runmode_relabel[self.mode]
                self.logger.debug('Set new run mode %s', self.mode)
                
                # set routine
                self.fit_routine_label['text'] = self.fitter.__name__
                
                # set run functions        
                fn_titles = self.fitter.function_names[self.mode]
                self.fit_function_title_box['values'] = fn_titles
                if self.fit_function_title.get() == '':
                    self.fit_function_title.set(fn_titles[0])
                    
        except UnboundLocalError:
            self.fit_function_title_box['values'] = ()
            self.fit_function_title.set("")
            self.fit_runmode_label['text'] = ""
            self.mode = ""
        
        # make shared_var dictionary
        self._make_shared_var_dict()
        
        # delete unused fitline objects
        for k in list(self.fit_lines.keys()):       # iterate fit list
            self.fit_lines[k].degrid()
            if k not in keylist:                    # check data list
                self.fit_lines_old[k] = self.fit_lines[k]
                del self.fit_lines[k]
        
        # make or regrid fitline objects
        n = 0
        for k in keylist:
            if k not in self.fit_lines.keys():
                if k in self.fit_lines_old.keys():
                    self.fit_lines[k] = self.fit_lines_old[k]
                else:
                    self.fit_lines[k] = fitline(self.bfit, self.runframe, dl[k], n)
            self.fit_lines[k].grid(n)
            n+=1
        
        self.populate_param()
            
    # ======================================================================= #
    def populate_param(self, *args, force_modify=False):
        """
            Populate the list of parameters
            
            force_modify: passed to line.populate
        """
        
        self.logger.debug('Populating fit parameters')
        
        # populate axis comboboxes
        lst = self.draw_components.copy()
        
        try:
            parlst = [p for p in self.fitter.gen_param_names(
                                                self.fit_function_title.get(), 
                                                self.n_component.get())]
        except KeyError:
            self.xaxis_combobox['values'] = []
            self.yaxis_combobox['values'] = []
            self.annotation_combobox['values'] = []
            return
        
        # Sort the parameters
        parlst.sort()
            
        # beta averaged T1
        if self.fit_function_title.get() == 'Str Exp':
            ncomp = self.n_component.get()
            
            if ncomp > 1: 
                for i in range(ncomp):
                    parlst.append('Beta-Avg 1/<T1>_%d' % i)
            else:
                parlst.append('Beta-Avg 1/<T1>')
            
        self.xaxis_combobox['values'] = [''] + parlst + lst
        self.yaxis_combobox['values'] = [''] + parlst + lst
        self.annotation_combobox['values'] = [''] + parlst + lst

        self._make_shared_var_dict()
            
        # turn off modify all so we don't cause an infinite loop
        modify_all_value = self.set_as_group.get()
        self.set_as_group.set(False)
            
        # regenerate fitlines
        for k in self.fit_lines.keys():
            self.fit_lines[k].populate(force_modify=force_modify)
        
        # reset modify all value
        self.set_as_group.set(modify_all_value)
            
    # ======================================================================= #
    def do_add_param(self, *args):
        """Launch popup for adding user-defined parameters to draw"""
        
        self.logger.info('Launching add paraemeter popup')
        
        if hasattr(self, 'pop_addpar'):
            p = self.pop_addpar
            
            # don't make more than one window
            if Toplevel.winfo_exists(p.win): 
                p.win.lift()
                return
            
            # make a new window, using old inputs and outputs
            self.pop_addpar = popup_add_param(self.bfit, 
                                    input_fn_text=p.input_fn_text)
        
        # make entirely new window
        else:
            self.pop_addpar = popup_add_param(self.bfit)
        
    # ======================================================================= #
    def do_fit(self, *args):
        # fitter
        fitter = self.fitter
        figstyle = 'fit'
        
        # get fitter inputs
        fn_name = self.fit_function_title.get()
        ncomp = self.n_component.get()
        
        xlims = [self.xlo.get(), self.xhi.get()]
        if not xlims[0]: 
            xlims[0] = '-inf'
            self.xlo.set('-inf')
        if not xlims[1]: 
            xlims[1] = 'inf'
            self.xhi.set('inf')
        
        try:
            xlims = tuple(map(float, xlims))
        except ValueError as err:
            messagebox.showerror("Error", 'Bad input for xlims')
            self.logger.exception(str(err))
            raise err
            
        self.logger.info('Fitting with "%s" with %d components', fn_name, ncomp)
        
        # build data list
        data_list = []
        for key in self.fit_lines:
            
            # get fit line
            fitline = self.fit_lines[key]
            
            # bdata object
            bdfit = fitline.dataline.bdfit
            
            # pdict
            pdict = {}
            for parname in fitline.parentry.keys():
                
                # get entry values
                pline = fitline.parentry[parname]
                line = []
                for col in fitline.collist:
                    
                    # get number entries
                    if col in ('p0', 'blo', 'bhi'):
                        try:
                            line.append(float(pline[col][0].get()))
                        except ValueError as errmsg:
                            self.logger.exception("Bad input.")
                            messagebox.showerror("Error", str(errmsg))
                    
                    # get "Fixed" entry
                    elif col in ['fixed']:
                        line.append(pline[col][0].get())
                
                    # get "Shared" entry
                    elif col in ['shared']:
                        line.append(pline[col][0].get())
                
                # make dict
                pdict[parname] = line
                
            # doptions
            doptions = {}
            
            if self.use_rebin.get():
                doptions['rebin'] = bdfit.rebin.get()
            
            if self.mode in ('1f', '1w'):
                dline = self.bfit.fetch_files.data_lines[key]
                doptions['omit'] = dline.bin_remove.get()
                if doptions['omit'] == dline.bin_remove_starter_line: 
                    doptions['omit'] = ''
            elif self.mode in ('20', '2h', '2e'):
                pass
            else:
                msg = 'Fitting mode %s not recognized' % self.mode
                self.logger.error(msg)
                raise RuntimeError(msg)
            
            # make data list
            data_list.append([bdfit, pdict, doptions])
        
        # call fitter with error message, potentially
        self.fit_input = (fn_name, ncomp, data_list)
        fit_status_window = self.make_fit_status_window()
        
        # do fit then kill window
        for d in data_list:
            self.logger.info('Fitting run %s: %s', self.bfit.get_run_key(d[0]), d[1:])    
        try:
            # fit_output keyed as {run:[key/par/cov/chi/fnpointer]}
            fit_output, gchi = fitter(fn_name=fn_name, 
                                      ncomp=ncomp, 
                                      data_list=data_list, 
                                      hist_select=self.bfit.hist_select, 
                                      asym_mode=self.bfit.get_asym_mode(self), 
                                      xlims=xlims)
        except Exception as errmsg:
            self.logger.exception('Fitting error')
            messagebox.showerror("Error", str(errmsg))
            raise errmsg from None
        finally:
            fit_status_window.destroy()
            del fit_status_window

        # set output results
        for key in fit_output.keys():
            self.bfit.data[key].set_fitresult(fit_output[key])
            self.bfit.data[key].fit_title = self.fit_function_title.get()
            self.bfit.data[key].ncomp = self.n_component.get()
            
        # display run results
        for key in self.fit_lines.keys():
            self.fit_lines[key].show_fit_result()
        
        # show global chi
        self.gchi_label['text'] = str(np.around(gchi, 2))
        
        self.do_end_of_fit()
        
    # ======================================================================= #
    def do_end_of_fit(self):
        """Things to do after fitting: draw, set checkbox status"""
        
        # enable fit checkboxes on fetch files tab
        for k in self.bfit.fetch_files.data_lines.keys():
            dline = self.bfit.fetch_files.data_lines[k]
            dline.draw_fit_checkbox['state'] = 'normal'
            dline.draw_res_checkbox['state'] = 'normal'
            dline.check_fit.set(True)
        self.bfit.fetch_files.check_state_fit.set(True)
        
        # change fetch asymmetry mode to match fit tab
        inv_map = {v: k for k, v in self.bfit.asym_dict.items()}
        asym_mode_fit = inv_map[self.bfit.get_asym_mode(self)]
        asym_mode_fetch = inv_map[self.bfit.get_asym_mode(self.bfit.fetch_files)]
        self.bfit.fetch_files.asym_type.set(asym_mode_fit)
        
        # draw fit results
        style = self.bfit.draw_style.get()
        
        if style in ['redraw', 'new']:
            self.bfit.draw_style.set('stack')
        
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.bfit.fetch_files.draw_all(figstyle='fit', ignore_check=False)
        
        if len(self.fit_lines.keys()) > self.bfit.legend_max_draw:
            
            try:
                self.plt.gca('fit').get_legend().remove()
            except AttributeError:
                pass
            else:
                self.plt.tight_layout('fit')
        
        # reset style and asym mode 
        self.bfit.draw_style.set(style)
        self.bfit.fetch_files.asym_type.set(asym_mode_fetch)
        
    # ======================================================================= #
    def do_fit_constraints(self):
        
        self.logger.info('Launching fit constraints popup')
        
        if hasattr(self, 'pop_fitconstr'):
            p = self.pop_fitconstr
            
            # don't make more than one window
            if Toplevel.winfo_exists(p.win): 
                p.win.lift()
                return
            
            # make a new window, using old inputs and outputs
            self.pop_fitconstr = popup_fit_constraints(self.bfit, 
                                    output_par_text=p.output_par_text_val, 
                                    output_text=p.output_text_val)
        
        # make entirely new window
        else:
            self.pop_fitconstr = popup_fit_constraints(self.bfit)
    
    # ======================================================================= #
    def do_fit_model(self):
        
        self.logger.info('Launching fit model popup')
        
        if hasattr(self, 'pop_fitres'):
            p = self.pop_fitres
            
            # don't make more than one window
            if Toplevel.winfo_exists(p.win): 
                p.win.lift()
                return
            
            # make a new window, using old inputs and outputs
            self.pop_fitres = popup_fit_results(self.bfit, 
                                    input_fn_text=p.input_fn_text, 
                                    output_par_text=p.output_par_text_val, 
                                    output_text=p.output_text_val, 
                                    chi=p.chi, 
                                    x = p.xaxis.get(), 
                                    y = p.yaxis.get())        
        
        # make entirely new window
        else:
            self.pop_fitres = popup_fit_results(self.bfit)
    
    # ======================================================================= #
    def do_gui_param(self, id=''):
        """Set initial parmeters with GUI"""
        
        self.logger.info('Launching initial fit parameters popup')
        popup_param(self.bfit, id)
        
    # ======================================================================= #
    def do_set_result_as_initial(self, *args):
        """Set initial parmeters as the fitting results"""
        
        self.logger.info('Setting initial parameters as fit results')
        
        # turn off modify all 
        modify_all_value = self.set_as_group.get()
        self.set_as_group.set(False)
        
        # set result to initial value
        for k in self.fit_lines.keys():
            
            # get line
            line = self.fit_lines[k]
            
            # get parameters
            parentry = line.parentry
            
            # set 
            for p in parentry.keys():
                parentry[p]['p0'][0].set(parentry[p]['res'][0].get()) 
    
        # reset modify all setting
        self.set_as_group.set(modify_all_value)
        
    # ======================================================================= #
    def do_reset_initial(self, *args):
        """Reset initial parmeters to defaults"""
                
        self.logger.info('Reset initial parameters')
                
        for k in self.fit_lines.keys():
            self.fit_lines[k].populate(force_modify=True)
    
    # ======================================================================= #
    def draw_residual(self, id, figstyle, rebin=1, **drawargs):
        """Draw fitting residuals for a single run"""
        
        self.logger.info('Drawing residual for run %s, rebin %d, '+\
                         'standardized: %s, %s', id, rebin, 
                         self.bfit.draw_standardized_res.get(), drawargs)
        
        # get draw setting 
        figstyle = 'data'
        draw_style = self.bfit.draw_style
        plt.ion()
        
        # get data and fit results
        data = self.bfit.data[id]
        fit_par = [data.fitpar['res'][p] for p in data.parnames]
        fn = data.fitfn
        data = data.bd
        
        # default label value
        if 'label' not in drawargs.keys():
            label = str(data.run)
        else:
            label = drawargs.pop('label', None)
            
        # set drawing style arguments
        for k in self.bfit.style:
            if k not in drawargs.keys():
                drawargs[k] = self.bfit.style[k]
        
        # make new window
        style = self.bfit.draw_style.get()
        if style == 'new' or not self.plt.active[figstyle]:
            self.plt.figure(figstyle)
        elif style == 'redraw':
            self.plt.clf(figstyle)
            
        ax = self.plt.gca(figstyle)
        ax.get_xaxis().get_major_formatter().set_useOffset(False)
        
        # get draw style
        style = self.bfit.draw_style.get()

        # get residuals
        x, a, da = data.asym(self.bfit.get_asym_mode(self), rebin=rebin)
        res = a - fn(x, *fit_par)
            
        # set x axis
        if data.mode in self.bfit.units: 
            unit = self.bfit.units[data.mode]
            x *= unit[0]
            xlabel = self.bfit.xlabel_dict[self.mode] % unit[1]
        else:
            xlabel = self.bfit.xlabel_dict[self.mode]

        # draw 
        if self.bfit.draw_standardized_res.get():
            self.plt.errorbar(figstyle, id, x, res/da, np.zeros(len(x)), 
                              label=label, **drawargs)
            
            # draw fill
            ax = self.plt.gca(figstyle)
            lim = ax.get_xlim()
            for i in range(1, 4):
                ax.fill_between(lim, -1*i, i, color='k', alpha=0.1)
            self.plt.xlim(figstyle, lim)
            self.plt.ylabel(figstyle, r'Standardized Residual ($\sigma$)')
        else:
            self.plt.errorbar(figstyle, id, x, res, da, label=label, **drawargs)
            self.plt.ylabel(figstyle, 'Residual')
        
        # draw pulse marker
        if '2' in data.mode: 
            self.plt.axvline(figstyle, 'line', data.get_pulse_s(), ls='--', color='k')
            unq = False
        else:
            unq = True
            
        # plot elements
        self.plt.xlabel(figstyle, xlabel)
        self.plt.axhline(figstyle, 'line', 0, color='k', linestyle='-', zorder=20, 
                        unique=unq)
        
        # show
        self.plt.tight_layout(figstyle)
        self.plt.legend(figstyle)
        
        raise_window()
        
    # ======================================================================= #
    def draw_fit(self, id, figstyle, unique=True, **drawargs):
        """
            Draw fit for a single run
            
            id: id of run to draw fit of 
            figstyle: one of "data", "fit", or "param" to choose which figure 
                    to draw in
        """
        
        self.logger.info('Drawing fit for run %s. %s', id, drawargs)
                     
        # get data and fit results
        data = self.bfit.data[id]
        fit_par = [data.fitpar['res'][p] for p in data.parnames]
        fn = data.fitfn
        
        # get draw style
        style = self.bfit.draw_style.get()
        
        # label reset
        if 'label' not in drawargs.keys():
            drawargs['label'] = self.bfit.data[id].label.get()
        drawargs['label'] += ' (fit)'
        label = drawargs['label']
        
        # set drawing style
        draw_id = data.id
        
        # make new window
        if style == 'new' or not self.plt.active[figstyle]:
            self.plt.figure(figstyle)
        elif style == 'redraw':
            self.plt.figure(figstyle)
            self.plt.clf(figstyle)
            
        # set drawing style arguments
        for k in self.bfit.style:
            if k not in drawargs.keys() \
                    and 'marker' not in k \
                    and k not in ['elinewidth', 'capsize']:
                drawargs[k] = self.bfit.style[k]
        
        # linestyle reset
        if drawargs['linestyle'] == 'None': 
            drawargs['linestyle'] = '-'
        
        # draw
        asym_mode = self.bfit.get_asym_mode(self)
        t, a, da = data.asym(asym_mode)
        
        fitx = np.linspace(min(t), max(t), self.n_fitx_pts)
        
        if self.mode in self.bfit.units: 
            unit = self.bfit.units[self.mode]
            fitxx = fitx*unit[0]
            xlabel = self.bfit.xlabel_dict[self.mode] % unit[1]
        else:                   
            fitxx = fitx
            xlabel = self.bfit.xlabel_dict[self.mode]
    
        # get fity
        fity = fn(fitx, *fit_par)
    
        # account for normalized draw modes
        draw_mode = self.bfit.asym_dict[self.bfit.fetch_files.asym_type.get()]
        if draw_mode == 'cn1':
            draw_mode += 'f'
            fity /= data.fitpar['res']['baseline']
        
        elif draw_mode == 'cn2':
            draw_mode += 'f'
            if 'amp' in data.fitpar['res'].keys():
                fity /= data.fitpar['res']['amp']
            else:
                fity /= fn(t[0], *par)
        
        elif draw_mode == 'cs':
            draw_mode += 'f'
            fity -= data.fitpar['res']['baseline']
                
        self.plt.plot(figstyle, draw_id, fitxx, fity, zorder=10, 
                      unique=unique, **drawargs)
        
        # plot elements
        self.plt.ylabel(figstyle, self.bfit.ylabel_dict.get(draw_mode, 'Asymmetry'))
        self.plt.xlabel(figstyle, xlabel)
        
        # show
        self.plt.tight_layout(figstyle)
        self.plt.legend(figstyle)
        
        # bring window to front 
        raise_window()
    
    # ======================================================================= #
    def draw_param(self, *args):
        """Draw the fit parameters"""
        figstyle = 'param'
        
        # make sure plot shows
        plt.ion()
        
        # get draw components
        xdraw = self.xaxis.get()
        ydraw = self.yaxis.get()
        ann = self.annotation.get()
        label = self.par_label.get()
        
        self.logger.info('Draw fit parameters "%s" vs "%s" with annotation "%s"'+\
                         ' and label %s', ydraw, xdraw, ann, label)
        
        # get plottable data
        try:
            xvals, xerrs = self.get_values(xdraw)
            yvals, yerrs = self.get_values(ydraw)
        except UnboundLocalError as err:
            self.logger.error('Bad input parameter selection')
            messagebox.showerror("Error", 'Select two input parameters')
            raise err from None
        except (KeyError, AttributeError) as err:
            self.logger.error('Parameter "%s or "%s" not found for drawing', 
                              xdraw, ydraw)
            messagebox.showerror("Error", 
                    'Drawing parameter "%s" or "%s" not found' % (xdraw, ydraw))
            raise err from None
        
        # get asymmetric errors
        if type(xerrs) is tuple:
            xerrs_l = xerrs[0]
            xerrs_h = xerrs[1]
        else:
            xerrs_l = xerrs
            xerrs_h = xerrs
        
        if type(yerrs) is tuple:
            yerrs_l = yerrs[0]
            yerrs_h = yerrs[1]
        else:
            yerrs_l = yerrs
            yerrs_h = yerrs
        
        # get annotation
        if ann != '':
            try:
                ann, _ = self.get_values(ann)
            except UnboundLocalError:
                ann = None
            except (KeyError, AttributeError) as err:
                self.logger.error('Bad input annotation value "%s"', ann)
                messagebox.showerror("Error", 
                        'Annotation "%s" not found' % (ann))
                raise err from None
        
        # fix annotation values (blank to none)
        else:
            ann = None
        
        # get mouseover annotation labels
        mouse_label, _ = self.get_values('Unique Id')
        
        # sort by x values 
        idx = np.argsort(xvals)
        xvals = np.asarray(xvals)[idx]
        yvals = np.asarray(yvals)[idx]
        
        xerrs_l = np.asarray(xerrs_l)[idx]
        yerrs_l = np.asarray(yerrs_l)[idx]
        xerrs_h = np.asarray(xerrs_h)[idx]
        yerrs_h = np.asarray(yerrs_h)[idx]
        
        if ann is not None:
            ann = np.asarray(ann)[idx]
            
        mouse_label = np.asarray(mouse_label)[idx]
        
        # fix annotation values (round floats)
        if ann is not None: 
            number_string = '%.'+'%df' % self.bfit.rounding
            for i, a in enumerate(ann):
                if type(a) in [float, np.float64]:
                    ann[i] = number_string % np.around(a, self.bfit.rounding)
            
        # get default data_id
        if label:   
            draw_id = label
        else:
            draw_id = ''
            
            if self.bfit.draw_style.get() == 'stack':
                ax = self.plt.gca(figstyle)
                
        # make new window
        style = self.bfit.draw_style.get()
        if style == 'new' or not self.plt.active[figstyle]:
            self.plt.figure(figstyle)
        elif style == 'redraw':
            self.plt.clf(figstyle)
        
        # get axis 
        ax = self.plt.gca(figstyle)
        
        # set dates axis
        if xdraw in ('Start Time', ): 
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%y/%m/%d (%H:%M)'))
            xvals = np.array([datetime.datetime.fromtimestamp(x) for x in xvals])
            xerrs = None
            ax.tick_params(axis='x', which='major', labelsize='x-small')
        else:
            try:
                ax.get_xaxis().get_major_formatter().set_useOffset(False)
            except AttributeError:
                pass
        
        if ydraw in ('Start Time', ):    
            ax.yaxis.set_major_formatter(mdates.DateFormatter('%y/%m/%d (%H:%M)'))
            yvals = mdates.epoch2num(yvals)
            yerrs = None
            ax.tick_params(axis='y', which='major', labelsize='x-small')
        else:
            try:
                ax.get_yaxis().get_major_formatter().set_useOffset(False)
            except AttributeError:
                pass
            
        # remove component label
        ncomp = self.n_component.get()
        xsuffix = ''
        ysuffix = ''
        if ncomp > 1:
        
            fn_params = self.fitter.gen_param_names(self.fit_function_title.get(), ncomp)
            
            if xdraw in fn_params or 'Beta-Avg 1/<T1>' in xdraw:
                spl = xdraw.split('_')
                xdraw = '_'.join(spl[:-1])
                xsuffix = ' [%s]' % spl[-1]
            
            if ydraw in fn_params or 'Beta-Avg 1/<T1>' in ydraw:
                spl = ydraw.split('_')
                ydraw = '_'.join(spl[:-1])
                ysuffix = ' [%s]' % spl[-1]
                
            
        # pretty labels
        xdraw = self.fitter.pretty_param.get(xdraw, xdraw)
        ydraw = self.fitter.pretty_param.get(ydraw, ydraw)
        
        # add suffix for multiple labels
        xdraw = xdraw + xsuffix
        ydraw = ydraw + ysuffix
        
        # attempt to insert units and scale
        unit_scale, unit = self.bfit.units.get(self.mode, [1, ''])
        
        if '%s' in xdraw:        
            xdraw = xdraw % unit
            xvals *= unit_scale
            xerrs_h *= unit_scale
            xerrs_l *= unit_scale
        if '%s' in ydraw:
            ydraw = ydraw % unit
            yvals *= unit_scale
            yerrs_h *= unit_scale
            yerrs_l *= unit_scale
        
        # check for nan errors
        if all(np.isnan(xerrs_h)): xerrs_h = None
        if all(np.isnan(xerrs_l)): xerrs_l = None
        if all(np.isnan(yerrs_h)): yerrs_h = None
        if all(np.isnan(yerrs_l)): yerrs_l = None
        
        if xerrs_h is None and xerrs_l is None:     xerr = None
        else:                                       xerr = (xerrs_l, xerrs_h)
        if yerrs_h is None and yerrs_l is None:     yerr = None
        else:                                       yerr = (yerrs_l, yerrs_h)
        
        # draw
        f = self.plt.errorbar(  figstyle, 
                                draw_id, 
                                xvals, 
                                yvals, 
                                xerr = xerr, 
                                yerr = yerr, 
                                label=draw_id, 
                                annot_label=mouse_label, 
                                **self.bfit.style)
        self._annotate(draw_id, xvals, yvals, ann, color=f[0].get_color(), unique=False)
        
        # format date x axis
        if xerrs is None:   self.plt.gcf(figstyle).autofmt_xdate()
        
        
        # plot elements
        self.plt.xlabel(figstyle, xdraw)
        self.plt.ylabel(figstyle, ydraw)
        self.plt.tight_layout(figstyle)
        
        if draw_id:
            self.plt.legend(figstyle, fontsize='x-small')
        
        # bring window to front
        raise_window()
        
    # ======================================================================= #
    def export(self, savetofile=True):
        """Export the fit parameter and file headers"""
        # get values and errors
        val = {}
        
        for v in self.xaxis_combobox['values']:
            if v == '': continue
            
            try:
                v2 = self.get_values(v) 
                
            # value not found
            except (KeyError, AttributeError):
                continue
            
            # if other error, don't crash but print the result
            except Exception: 
                traceback.print_exc()
            else:
                val[v] = v2[0]
                
                if type(v2[1]) is tuple:                     
                    val['Error- '+v] = v2[1][0]
                    val['Error+ '+v] = v2[1][1]
                else:
                    val['Error '+v] = v2[1]
        
        # make data frame for output
        df = pd.DataFrame(val)
        df.set_index('Run Number', inplace=True)
        
        # drop completely empty columns
        bad_cols = [c for c in df.columns if all(df[c].isna())]
        for c in bad_cols:
            df.drop(c, axis='columns', inplace=True)
        
        if savetofile:
            
            # get file name
            filename = filedialog.asksaveasfilename(filetypes=[('csv', '*.csv'), 
                                                               ('allfiles', '*')], 
                                                defaultextension='.csv')
            if not filename:    return
            self.logger.info('Exporting parameters to "%s"', filename)
            
            # check extension 
            if os.path.splitext(filename)[1] == '':
                filename += '.csv'
            
            # write header
            data = self.bfit.data[list(self.fit_lines.keys())[0]]
            header = ['# Fit function : %s' % data.fit_title, 
                      '# Number of components: %d' % data.ncomp, 
                      '# Global Chi-Squared: %s' % self.gchi_label['text'], 
                      '#\n#\n']
            
            with open(filename, 'w') as fid:
                fid.write('\n'.join(header))
            
            # write data
            df.to_csv(filename, mode='a+')
            self.logger.debug('Export success')
        else:
            self.logger.info('Returned exported parameters')
            return df
        
    # ======================================================================= #
    def export_fit(self, savetofile=True):
        """Export the fit lines as csv files"""
        
        # filename
        filename = self.bfit.fileviewer.default_export_filename
        filename = '_fit'.join(os.path.splitext(filename))
        
        if not filename: return
        
        try:
            filename = filedialog.askdirectory()+'/'+filename
        except TypeError:
            pass
        
        # asymmetry type
        asym_mode = self.bfit.get_asym_mode(self)
        
        # get data and write
        for id in self.fit_lines.keys():
            
            # get data
            data = self.bfit.data[id]
            t, a, da = data.asym(asym_mode)
            
            # get fit data
            fitx = np.linspace(min(t), max(t), self.n_fitx_pts)
            
            try:
                fit_par = [data.fitpar['res'][p] for p in data.parnames]
            except AttributeError:
                continue
            dfit_par_l = [data.fitpar['dres-'][p] for p in data.parnames]
            dfit_par_h = [data.fitpar['dres+'][p] for p in data.parnames]
            fity = data.fitfn(fitx, *fit_par)
            
            if data.mode in self.bfit.units: 
                unit = self.bfit.units[data.mode]
                fitxx = fitx*unit[0]
                xlabel = self.bfit.xlabel_dict[self.mode] % unit[1]
            else:                   
                fitxx = fitx
                xlabel = self.bfit.xlabel_dict[self.mode]

            # write header
            fname = filename%(data.year, data.run)
            header = ['# %s' % data.id, 
                      '# %s' % data.title, 
                      '# Fit function : %s' % data.fit_title, 
                      '# Number of components: %d' % data.ncomp, 
                      '# Rebin: %d' % data.rebin.get(), 
                      '# Bin Omission: %s' % data.omit.get().replace(
                                self.bfit.fetch_files.bin_remove_starter_line, ''), 
                      '# Chi-Squared: %f' % data.chi, 
                      '# Parameter names: %s' % ', '.join(data.parnames), 
                      '# Parameter values: %s' % ', '.join(list(map(str, fit_par))), 
                      '# Parameter errors (-): %s' % ', '.join(list(map(str, dfit_par_l))), 
                      '# Parameter errors (+): %s' % ', '.join(list(map(str, dfit_par_h))), 
                      '#', 
                      '# Generated by bfit v%s on %s' % (__version__, datetime.datetime.now()), 
                      '#']
            
            with open(fname, 'w') as fid:
                fid.write('\n'.join(header) + '\n')
            
            # write data
            df = pd.DataFrame({xlabel:fitx, 'asymmetry':fity})
            df.to_csv(fname, index=False, mode='a+')
            self.logger.info('Exporting fit to %s', fname)
    
    # ======================================================================= #
    def get_values(self, select):
        """ Get plottable values"""
        
        data = self.bfit.data
        dlines = self.bfit.fetch_files.data_lines
        
        # draw only selected runs
        runs = [dlines[k].id for k in dlines if dlines[k].check_state.get()]
        runs.sort()
        
        self.logger.debug('Fetching parameter %s', select)
        
        # parameter names 
        parnames = self.fitter.gen_param_names(self.fit_function_title.get(), 
                                               self.n_component.get())
        
        # Data file options
        if select == 'Temperature (K)':
            val = [data[r].temperature.mean for r in runs]
            err = [data[r].temperature.std for r in runs]
        
        elif select == 'B0 Field (T)':
            val = [data[r].field for r in runs]
            err = [data[r].field_std for r in runs]
        
        elif select == 'RF Level DAC':
            try:
                val = [data[r].camp.rf_dac.mean for r in runs]
                err = [data[r].camp.rf_dac.std for r in runs]
            except AttributeError:
                pass
        
        elif select == 'Platform Bias (kV)':
            try:
                val = [data[r].bias for r in runs]
                err = [data[r].bias_std for r in runs]
            except AttributeError:
                pass
                
        elif select == 'Impl. Energy (keV)':
            val =  [data[r].bd.beam_kev() for r in runs]
            err =  [data[r].bd.beam_kev(get_error=True) for r in runs]
        
        elif select == 'Run Duration (s)':
            val = [data[r].bd.duration for r in runs]
            err = [np.nan for r in runs]
        
        elif select == 'Run Number':
            val = [data[r].run for r in runs]
            err = [np.nan for r in runs]
        
        elif select == 'Sample':
            val = [data[r].bd.sample for r in runs]
            err = [np.nan for r in runs]
            
        elif select == 'Start Time':
            val = [data[r].bd.start_time for r in runs]
            err = [np.nan for r in runs]
        
        elif select == 'Title':
            val = [data[r].bd.title for r in runs]
            err = [np.nan for r in runs]
        
        elif select == '1000/T (1/K)':
            val = [1000/data[r].temperature.mean for r in runs]
            err = [1000*data[r].temperature.std/(data[r].temperature.mean**2) \
                   for r in runs]
        
        elif select == 'Chi-Squared':
            val = []
            for r in runs:
                try:
                    val.append(data[r].chi)
                except(KeyError, AttributeError):
                    val.append(np.nan)
            err = [np.nan for r in runs]
        
        elif select == 'Year':
            val = [data[r].year for r in runs]
            err = [np.nan for r in runs]
        
        elif select == 'Unique Id':
            val = [data[r].id for r in runs]
            err = [np.nan for r in runs]

        elif 'Beta-Avg 1/<T1' in select:
            
            # get component
            idx = select.find('_')
            if idx < 0:     comp_num = ''
            else:           comp_num = select[idx:]
            comp_num = comp_num.replace('>', '')
            
            # initialize
            val = []
            err = []
                
            # get T1 and beta from that component average
            for r in runs:
                T1i = data[r].fitpar['res']['1_T1'+comp_num]
                T1 = 1/T1i
                dT1_l = data[r].fitpar['dres-']['1_T1'+comp_num]/(T1i**2)
                dT1_u = data[r].fitpar['dres+']['1_T1'+comp_num]/(T1i**2)
                
                dT1 = np.sqrt(np.square(dT1_l) + np.square(dT1_u))
                
                beta = data[r].fitpar['res']['beta'+comp_num]
                dbeta_l = data[r].fitpar['dres-']['beta'+comp_num]
                dbeta_u = data[r].fitpar['dres+']['beta'+comp_num]
                
                dbeta = np.sqrt(np.square(dbeta_l) + np.square(dbeta_u))
                
                # take average
                betai = 1./beta
                pd_T1 = gamma(betai)/beta
                pd_beta = -T1*gamma(betai)*(1+betai*polygamma(0, betai))*(betai**2)
                T1avg = T1*pd_T1
                dT1avg = ( (pd_T1*dT1)**2 + (pd_beta*dbeta)**2 )**0.5
                
                val.append(1/T1avg)
                err.append(dT1avg/(T1avg**2))

        elif 'Cryo Lift Set (mm)' in select:
            val = [data[r].bd.camp.clift_set.mean for r in runs]
            err = [data[r].bd.camp.clift_set.std for r in runs]
        
        elif 'Cryo Lift Read (mm)' in select:
            val = [data[r].bd.camp.clift_read.mean for r in runs]
            err = [data[r].bd.camp.clift_read.std for r in runs]
        
        elif 'He Mass Flow' in select:
            var = 'mass_read' if data[runs[0]].area == 'BNMR' else 'he_read'
            val = [data[r].bd.camp[var].mean for r in runs]
            err = [data[r].bd.camp[var].std for r in runs]
            
        elif 'CryoEx Mass Flow' in select:
            val = [data[r].bd.camp.cryo_read.mean for r in runs]
            err = [data[r].bd.camp.cryo_read.std for r in runs]
            
        elif 'Needle Set (turns)' in select:
            val = [data[r].bd.camp.needle_set.mean for r in runs]
            err = [data[r].bd.camp.needle_set.std for r in runs]
            
        elif 'Needle Read (turns)' in select:
            val = [data[r].bd.camp.needle_pos.mean for r in runs]
            err = [data[r].bd.camp.needle_pos.std for r in runs]
            
        elif 'Laser Power' in select:
            val = [data[r].bd.epics.las_pwr.mean for r in runs]
            err = [data[r].bd.epics.las_pwr.std for r in runs]
            
        elif 'Target Bias (kV)' in select:
            val = [data[r].bd.epics.target_bias.mean for r in runs]
            err = [data[r].bd.epics.target_bias.std for r in runs]        
        
        elif 'NBM Rate (count/s)' in select:
            rate = lambda b : np.sum([b.hist['NBM'+h].data \
                                    for h in ('F+', 'F-', 'B-', 'B+')])/b.duration
            val = [rate(data[r].bd) for r in runs]
            err = np.full(len(val), np.nan)
            
        elif 'Sample Rate (count/s)' in select:
            hist = ('F+', 'F-', 'B-', 'B+') if data[runs[0]].area == 'BNMR' \
                                         else ('L+', 'L-', 'R-', 'R+')
                
            rate = lambda b : np.sum([b.hist[h].data for h in hist])/b.duration
            val = [rate(data[r].bd) for r in runs]
            err = np.full(len(val), np.nan)     
            
        # fitted parameter options
        elif select in parnames:
            val = []
            err_l = []
            err_u = []
            
            for r in runs:
                try:
                    val.append(data[r].fitpar['res'][select])
                    err_l.append(data[r].fitpar['dres-'][select])
                    err_u.append(data[r].fitpar['dres+'][select])
                except KeyError:
                    val.append(np.nan)
                    err_l.append(np.nan)
                    err_u.append(np.nan)
            err = (err_l, err_u)
            
        # check user-defined parameters
        elif hasattr(self, 'pop_addpar') and select in self.pop_addpar.set_par.keys():
            val = self.pop_addpar.set_par[select]()
            err = np.full(len(val), np.nan)
        
        try:
            return (val, err)
        except UnboundLocalError:
            self.logger.warning('Parameter selection "%s" not found' % select)
            raise AttributeError('Selection "%s" not found' % select) from None
    
    # ======================================================================= #
    def load_state(self):
        """
            Load the state of the gui
        """
        
        # get the filename 
        filename = filedialog.askopenfilename(filetypes=[('yaml', '*.yaml'), 
                                                         ('allfiles', '*')])
        if not filename:
            return
        
        self.logger.info('Loading program state from %s', filename)
        
        # load the object with the data
        with open(filename, 'r') as fid:
            from_file = yaml.safe_load(fid)
    
        # clear loaded runs
        fetch_tab = self.bfit.fetch_files
        fetch_tab.remove_all()
    
        # load selected runs
        datalines = from_file['datalines']
        setyear = fetch_tab.year.get()
        setrun =  fetch_tab.run.get()
        for id in datalines:
            d = datalines[id]
            
            # set year and run and fetch
            fetch_tab.year.set(d['year'])
            fetch_tab.run.set(d['run'])
            fetch_tab.get_data()
            
            # set corresponding parameters for the run 
            d_actual = fetch_tab.data_lines[id]
            d_actual.bin_remove.set(d['bin_remove'])
            d_actual.check_data.set(d['check_data'])
            d_actual.check_fit.set(d['check_fit'])
            d_actual.check_res.set(d['check_res'])
            d_actual.check_state.set(d['check_state'])
            d_actual.label.set(d['label'])
            d_actual.rebin.set(d['rebin'])
    
        # reset year and run input info
        fetch_tab.year.set(setyear)
        fetch_tab.run.set(setrun)
        
        # set the fitting function 
        self.fit_function_title_box.set(from_file['fitfn'])
        
        # set the number of components
        self.n_component.set(from_file['ncomponents'])
        
        # set the global chisquared
        self.gchi_label['text'] = from_file['gchi']
        
        # set probe
        self.bfit.probe_species.set(from_file['probe_species'])
        self.bfit.set_probe_species()
        
        # get parameters in fitting page
        self.populate()
        
        # set parameter values
        d_fitdata = self.bfit.data
        fitlines = from_file['fitlines']
        for id in fitlines:
            parentry = fitlines[id]
            parentry_actual = self.fit_lines[id].parentry
            for parname in parentry:
                par = parentry[parname]
                for k in par.keys():
                    parentry_actual[parname][k][0].set(par[k])
                    
            # make sure dataline checkboxes are active
            fetch_tab.data_lines[id].draw_fit_checkbox['state'] = 'normal'
            fetch_tab.data_lines[id].draw_res_checkbox['state'] = 'normal'
        
            # set fit inputs
            d_fitdata[id].set_fitpar({p:[parentry[p]['p0'], 
                                         parentry[p]['blo'], 
                                         parentry[p]['bhi'], 
                                         parentry[p]['fixed']] for p in parentry})
            
            # get chisq
            keylist = self.fitter.gen_param_names(from_file['fitfn'], 
                                                  from_file['ncomponents'])
            for k in keylist: 
                if 'chi' in parentry[k].keys():
                    chi = float(parentry[k]['chi'])
                    break
                    
            # get pulse length
            d_actual = fetch_tab.data_lines[id]
            pulse_len = 0
            if d_actual.mode in ('20', '2h'):
                pulse_len = d_actual.bdfit.get_pulse_s()
                    
            # get probe lifetime
            lifetime = bd.life[from_file['probe_species']]
            
            # get fit function
            fitfn = self.fitter.get_fn(from_file['fitfn'], 
                                       from_file['ncomponents'], 
                                       pulse_len, 
                                       lifetime)
            
            if '2' in d_actual.mode and from_file['probe_species'] == 'Mg31':
                fitfn1 = lambda x, *par : fa_31Mg(x, pulse_len)*fitfn(x, *par)
            else:
                fitfn1 = fitfn

            # set fit results
            d_fitdata[id].set_fitresult([keylist, 
                              [float(parentry[p]['res']) for p in keylist], 
                              [float(parentry[p]['dres-']) for p in keylist], 
                              [float(parentry[p]['dres+']) for p in keylist], 
                              chi, 
                              fitfn1]
                            )
        
        # xlims
        self.xlo.set(from_file['xlo'])
        self.xhi.set(from_file['xhi'])
        
        # set minimizer
        self.bfit.minimizer.set(from_file['minimizer'])
        self.bfit.set_fit_routine()
        
        # set menu options
        self.bfit.norm_with_param.set(from_file['norm_with_param'])
        self.bfit.draw_standardized_res.set(from_file['draw_standardized_res'])
        self.bfit.use_nbm.set(from_file['use_nbm'])
        self.bfit.draw_ppm.set(from_file['draw_ppm'])
        self.bfit.thermo_channel.set(from_file['thermo_channel'])
        self.bfit.units = from_file['units']
        self.bfit.label_default.set(from_file['label_default'])
        self.bfit.ppm_reference = from_file['ppm_reference']
        self.bfit.update_period = from_file['update_period']
        
    # ======================================================================= #
    def make_fit_status_window(self):
        """Window to show that fitting is in progress"""
        fit_status_window = Toplevel(self.bfit.root)
        fit_status_window.lift()
        fit_status_window.resizable(FALSE, FALSE)
        ttk.Label(fit_status_window, 
                  text="Fitting in progress\nTo cancel press <Ctrl-C> in terminal ONCE", 
                  justify='center', 
                  pad=0).grid(column=0, row=0, padx=15, pady=15)
        fit_status_window.update_idletasks()
        self.bfit.root.update_idletasks()
        
        width = fit_status_window.winfo_reqwidth()
        height = fit_status_window.winfo_reqheight()
        
        rt_x = self.bfit.root.winfo_x()
        rt_y = self.bfit.root.winfo_y()
        rt_w = self.bfit.root.winfo_width()
        rt_h = self.bfit.root.winfo_height()
        
        x = rt_x + rt_w/2 - (width/2)
        y = rt_y + rt_h/3 - (width/2)
        
        fit_status_window.geometry('{}x{}+{}+{}'.format(width, height, int(x), int(y)))
        fit_status_window.update_idletasks()
        return fit_status_window
        
    # ======================================================================= #
    def modify_all(self, *args, source=None, par='', column=''):
        """
            Modify all input fields of each line to match the altered one, 
            conditional on self.set_as_group
            
            source_line: the fitline to copy
            parameter:   name of the parameter to copy
            column:      name of the column to copy            
        """
        
        setall = self.set_as_group.get()
        self.logger.info('Set modify all as %s', setall)
        
        for k in self.fit_lines.keys():
            self.fit_lines[k].set_input(source, par, column, setall)        
    
    # ======================================================================= #
    def return_binder(self):
        """
            Binding to entery key press, depending on focus. 
            
            FOCUS                   ACTION
            
            comboboxes or buttons   draw_param
                in right frame
            else                    do_fit
        """
    
        # get focus
        focus = self.bfit.root.focus_get()
        
        # right frame items
        draw_par_items = (  self.xaxis_combobox, 
                            self.yaxis_combobox, 
                            self.annotation_combobox, 
                            self.par_label_entry)
        
        # do action 
        if focus in draw_par_items:
            self.draw_param()
        elif focus == self.n_component_box:
            self.populate_param(force_modify=True)
        elif focus == self.bfit.root:
            pass
        else:
            self.do_fit()
    
    # ======================================================================= #
    def save_state(self):
        """
            Save the state of the gui:
            
            dataline state info
            Fitting function
            Number of components 
            Initial inputs
            Fit results
        """
        
        # final output 
        to_file = {}
        
        # get state from datalines
        datalines = self.bfit.fetch_files.data_lines
        dlines = {}
        for id in datalines:
            d = datalines[id]
            dlines[id] = {
                    'bin_remove'   :d.bin_remove.get(), 
                    'check_data'   :d.check_data.get(), 
                    'check_fit'    :d.check_fit.get(), 
                    'check_res'    :d.check_res.get(), 
                    'check_state'  :d.check_state.get(), 
                    'id'           :d.id, 
                    'label'        :d.label.get(), 
                    'rebin'        :d.rebin.get(), 
                    'run'          :d.run, 
                    'year'         :d.year
                    }
        to_file['datalines'] = dlines
        
        # get state of fitting info from fit page
        to_file['fitfn'] = self.fit_function_title.get()
        to_file['ncomponents'] = self.n_component.get()
        to_file['gchi'] = self.gchi_label['text']
        to_file['probe_species'] = self.bfit.probe_species.get()
        to_file['minimizer'] = self.bfit.minimizer.get()
        to_file['norm_with_param'] = self.bfit.norm_with_param.get()
        to_file['draw_standardized_res'] = self.bfit.draw_standardized_res.get()
        to_file['use_nbm'] = self.bfit.use_nbm.get()
        to_file['draw_ppm'] = self.bfit.draw_ppm.get()
        to_file['thermo_channel'] = self.bfit.thermo_channel.get()
        to_file['units'] = self.bfit.units
        to_file['label_default'] = self.bfit.label_default.get()
        to_file['ppm_reference'] = self.bfit.ppm_reference
        to_file['update_period'] = self.bfit.update_period
        
        # get parameter values from fitlines
        fitlines = self.fit_lines
        flines = {}
        for id in fitlines:
            parentry_actual = fitlines[id].parentry
            parentry = {}
            for param_name in parentry_actual:
                par = parentry_actual[param_name]
                parentry[param_name] = {k:par[k][0].get() for k in par}
            flines[id] = parentry
        to_file['fitlines'] = flines
  
        # get xlims
        to_file['xlo'] = self.xlo.get()
        to_file['xhi'] = self.xhi.get()
  
        # save file ----------------------------------------------------------
        fid = filedialog.asksaveasfile(mode='w', filetypes=[('yaml', '*.yaml'), 
                                                           ('allfiles', '*')], 
                                       defaultextension='.yaml')
        if fid:
            yaml.dump(to_file, fid)
            fid.close()
    
        self.logger.info('Saving program state to %s', fid)
    
    # ======================================================================= #
    def show_all_results(self):
        """Make a window to display table of fit results"""
        
        self.logger.info('Launching parameter table popup')
        
        # get fit results
        df = self.export(savetofile=False)
        popup_show_param(df)
        
    # ======================================================================= #
    def update_param(self, *args):
        """Update all figures with parameters drawn with new fit results"""
        
        self.logger.info('Updating parameter figures')
        
        # get list of figure numbers for parameters
        figlist = self.plt.plots['param']
        
        # set style to redraw
        current_active = self.plt.active['param']
        current_style = self.bfit.draw_style.get()
        self.bfit.draw_style.set('stack')
        
        # get current labels
        current_xlab = self.xaxis.get()
        current_ylab = self.yaxis.get()
        
        for fig_num in figlist:
            
            # get figure and drawn axes
            ax = plt.figure(fig_num).axes[0]
            xlab = ax.get_xlabel()
            ylab = ax.get_ylabel()
            xscale = ax.get_xscale()
            yscale = ax.get_yscale()
            
            # back-translate pretty labels to originals
            ivd = {v: k for k, v in self.fitter.pretty_param.items()}
            xlab = ivd.get(xlab, xlab)
            ylab = ivd.get(ylab, ylab)
            
            # set new labels
            self.xaxis.set(xlab)
            self.yaxis.set(ylab)
            
            # draw new labels
            self.plt.active['param'] = fig_num
            self.draw_param()
            
            # scale the plot
            ax = plt.figure(fig_num).axes[0]
            ax.set_xscale(xscale)
            ax.set_yscale(yscale)
            
            self.logger.debug('Updated figure %d (%s vs %s)', fig_num, ylab, xlab)
            
        # reset to old settings
        self.bfit.draw_style.set(current_style)
        self.xaxis.set(current_xlab)
        self.yaxis.set(current_ylab)
        self.plt.active['param'] = current_active
        plt.figure(current_active)
        
# =========================================================================== #
# =========================================================================== #
class fitline(object):
    """
        Instance variables 
        
            bfit            pointer to top class
            dataline        pointer to dataline object in fetch_files_tab
            disable_entry_callback  disables copy of entry strings to 
                                    dataline.bdfit parameter values
            parent          pointer to parent object (frame)
            parlabels       label objects, saved for later destruction
            parentry        [parname][colname] of ttk.Entry objects saved for 
                            retrieval and destruction
            run_label       label for showing which run is selected
            run_label_title label for showing which run is selected
            fitframe        mainframe for this tab. 
    """
    
    n_runs_max = 5      # number of runs before scrollbar appears
    collist = ['p0', 'blo', 'bhi', 'res', 'dres-', 'dres+', 'chi', 'fixed', 'shared']
    selected = 0        # index of selected run 
    
    # ======================================================================= #
    def __init__(self, bfit, parent, dataline, row):
        """
            Inputs:
                bfit:       top level pointer
                parent:     pointer to parent frame object
                dataline:   fetch_files.dataline object corresponding to the 
                                data we want to fit
                row:        grid position
        """
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.debug('Initializing fit line for run %d in row %d', 
                          dataline.run, row)
        
        # initialize
        self.bfit = bfit
        self.parent = parent
        self.dataline = dataline
        self.row = row
        self.parlabels = []
        self.parentry = {}
        self.disable_entry_callback = False
             
        # get parent frame
        fitframe = ttk.Frame(self.parent, pad=(5, 0))
        
        # label for displyaing run number
        if type(self.dataline.bdfit.bd) is bdata:
            self.run_label = Label(fitframe, 
                            text='[ %d - %d ]' % (self.dataline.run, 
                                                  self.dataline.year), 
                           bg=colors.foreground, fg=colors.background)
            
        elif type(self.dataline.bdfit.bd) is bmerged:
            runs = textwrap.wrap(str(self.dataline.run), 5)
             
            self.run_label = Label(fitframe, 
                                text='[ %s ]' % ' + '.join(runs), 
                                bg=colors.foreground, fg=colors.background)

        # title of run
        self.run_label_title = Label(fitframe, 
                            text=self.dataline.bdfit.title, 
                            justify='right', fg='red3')
                    
        # Parameter input labels
        gui_param_button = ttk.Button(fitframe, text='Initial Value', 
                        command=lambda : self.bfit.fit_files.do_gui_param(id=self.dataline.id), 
                        pad=0)
        result_comp_button = ttk.Button(fitframe, text='Result', 
                        command=self.show_fn_composition, pad=0)
        
        c = 0
        ttk.Label(fitframe, text='Parameter').grid(    column=c, row=1, padx=5); c+=1
        gui_param_button.grid(column=c, row=1, padx=5, pady=2); c+=1
        ttk.Label(fitframe, text='Low Bound').grid(    column=c, row=1, padx=5); c+=1
        ttk.Label(fitframe, text='High Bound').grid(   column=c, row=1, padx=5); c+=1
        result_comp_button.grid(column=c, row=1, padx=5, pady=2, sticky=(E, W)); c+=1
        ttk.Label(fitframe, text='Error (-)').grid(        column=c, row=1, padx=5); c+=1
        ttk.Label(fitframe, text='Error (+)').grid(        column=c, row=1, padx=5); c+=1
        ttk.Label(fitframe, text='ChiSq').grid(        column=c, row=1, padx=5); c+=1
        ttk.Label(fitframe, text='Fixed').grid(        column=c, row=1, padx=5); c+=1
        ttk.Label(fitframe, text='Shared').grid(       column=c, row=1, padx=5); c+=1
    
        self.run_label.grid(column=0, row=0, padx=5, pady=5, columnspan=2, sticky=W)
        self.run_label_title.grid(column=1, row=0, padx=5, pady=5, columnspan=c-1, sticky=E)
        
        # save frame 
        self.fitframe = fitframe
        
        # resizing
        for i in range(c):
            self.fitframe.grid_columnconfigure(i, weight=1)
        
        # fill with initial parameters
        self.parlabels = []     # track all labels and inputs
        self.populate()
                
    # ======================================================================= #
    def __del__(self):
        
        if hasattr(self, 'parlabels'):   del self.parlabels
        
        # kill buttons and frame
        try:
            for child in self.parent.winfo_children():
                child.destroy()
        except Exception:
            pass
    
        if hasattr(self, 'parentry'):    del self.parentry
    
    # ======================================================================= #
    def populate(self, force_modify=False):
        """
            Fill and grid new parameters. Reuse old fields if possible
            
            force_modify: if true, clear and reset parameter inputs. 
        """
        
        # get list of parameters and initial values
        try:
            plist = self.get_new_parameters()
        except KeyError as err:
            return          # returns if no parameters found
        except RuntimeError as err:
            messagebox.showerror('RuntimeError', err)
            raise err from None
        else:
            n_old_par = len(self.parlabels)
            n_new_par = len(plist)
            min_n_par = min(n_old_par, n_new_par)
            parkeys = list(self.parentry.keys())    # old parameter keys
            parkeys.sort()
            
            # destroy excess labels and entries
            for i in range(n_new_par, n_old_par):
                self.parlabels[-1].destroy()
                for p in self.parentry[parkeys[i]].keys():
                    self.parentry[parkeys[i]][p][1].destroy()
                    
                del self.parlabels[-1]
                del self.parentry[parkeys[i]]
        
        self.logger.debug('Populating parameter list with %s', plist)

        # get data and frame
        fitframe = self.fitframe
        fitdat = self.dataline.bdfit
    
        # labels ------------------------------------------------------------
        c = 0

        # repurpose old labels
        for i in range(min_n_par):
            self.parlabels[i]['text'] = plist[i]
        
        # make new labels
        for i in range(n_old_par, n_new_par):
            self.parlabels.append(ttk.Label(fitframe, text=plist[i], justify=LEFT))
            self.parlabels[-1].grid(column=c, row=2+i, padx=5, sticky=E)
        
        # move all parameters entries and values to new key set
        new_parentry = {}
        for i in range(min_n_par):
            p = plist[i]
            p_old = parkeys[i]
            new_parentry[p] = self.parentry[p_old]
        self.parentry = new_parentry
        
        # initial parameters ------------------------------------------------- 
        
        # repurpose old parameter fields
        r = 1
        self.disable_entry_callback = True  # prevent setting bad data
        for i in range(min_n_par):
            p = plist[i]
            c = 1
            r += 1
            
            # clear entry and insert new text
            for col in ('p0', 'blo', 'bhi'):                
                entry = self.parentry[p][col][1]
                
                if force_modify:
                    entry.delete(0, 'end')
                    self.parentry[p]['fixed'][0].set(fitdat.fitpar['fixed'][p])
                    
                if not entry.get():
                    entry.insert(0, str(fitdat.fitpar[col][p]))
                
                entry.grid(column=c, row=r, padx=5, sticky=E); c += 1
            
        r = min_n_par+1
        
        self.disable_entry_callback = False
                 
        # make new parameter fields
        for i in range(n_old_par, n_new_par):
            p = plist[i]            
            self.parentry[p] = {}
            
            c = 0               # gridding column         
            r += 1              # gridding row         
            
            for col in ('p0', 'blo', 'bhi'):
                c += 1
                value = StringVar()
                entry = ttk.Entry(fitframe, textvariable=value, width=13)
                entry.insert(0, str(fitdat.fitpar[col][p]))
                entry.grid(column=c, row=r, padx=5, sticky=E)
                self.parentry[p][col] = (value, entry)
        
        # fit results ------------------------------------------------------- 
        
        # repurpose old result fields
        r = 1
        for i in range(min_n_par):
            r += 1
            c = 4
            p = plist[i]
            
            # clear text in parentry fields
            for col in ('res', 'dres-', 'dres+', 'chi'):
                if col in self.parentry[p].keys():  # exception needed for chi
                    par = self.parentry[p][col][1]
                    par.delete(0, 'end')
                    
                    if col == 'chi':
                        par.grid(column=c, row=r, padx=5, sticky=E, rowspan=len(plist))
                    else:
                        par.grid(column=c, row=r, padx=5, sticky=E)
                c += 1
                    
            # do fixed box
            self.parentry[p]['fixed'][1].grid(column=c, row=r, padx=5, sticky=E); c += 1
            
            # do shared box
            self.parentry[p]['shared'][1].grid(column=c, row=r, padx=5, sticky=E); c += 1
        
        # make new result fields
        r = min_n_par+1
        for i in range(n_old_par, n_new_par):
            r += 1
            c = 4
            p = plist[i]
            
            # do results
            par_val = StringVar()
            par = ttk.Entry(fitframe, textvariable=par_val, width=15)
            par['state'] = 'readonly'
            par['foreground'] = colors.foreground
            
            dpar_val_l = StringVar()
            dpar_val_u = StringVar()
            dpar_l = ttk.Entry(fitframe, textvariable=dpar_val_l, width=15)
            dpar_u = ttk.Entry(fitframe, textvariable=dpar_val_u, width=15)
            dpar_l['state'] = 'readonly'
            dpar_l['foreground'] = colors.foreground
            dpar_u['state'] = 'readonly'
            dpar_u['foreground'] = colors.foreground
                                     
            par. grid(column=c, row=r, padx=5, sticky=E); c += 1
            dpar_l.grid(column=c, row=r, padx=5, sticky=E); c += 1
            dpar_u.grid(column=c, row=r, padx=5, sticky=E); c += 1
            
            # do chi only once
            if i==0:
                chi_val = StringVar()
                chi = Entry(fitframe, textvariable=chi_val, width=7)
                chi['state'] = 'readonly'
                chi['foreground'] = colors.foreground
                
                chi.grid(column=c, row=r, padx=5, sticky=E, rowspan=len(plist)); 
                self.parentry[p]['chi'] = (chi_val, chi)
            c += 1
            
            # save ttk.Entry objects in dictionary [parname][colname]
            self.parentry[p]['res'] = (par_val, par)
            self.parentry[p]['dres-'] = (dpar_val_l, dpar_l)
            self.parentry[p]['dres+'] = (dpar_val_u, dpar_u)
            
            # do fixed box
            value = BooleanVar()
            entry = ttk.Checkbutton(fitframe, text='', \
                                     variable=value, onvalue=True, offvalue=False)
            entry.grid(column=c, row=r, padx=5, sticky=E); c += 1
            self.parentry[p]['fixed'] = (value, entry)
            value.set(fitdat.fitpar['fixed'][p])
            
            # do shared box
            entry = ttk.Checkbutton(fitframe, text='', onvalue=True, offvalue=False)
            entry.grid(column=c, row=r, padx=5, sticky=E); c += 1
            self.parentry[p]['shared'] = [value, entry]
            
        # set p0 synchronization ----------------------------------------------
        
        # make callback function to set p0 values in bdfit object
        def callback(*args, parname, col, source):
            
            if self.disable_entry_callback:
                return 
                
            # set parameter entry synchronization
            self.bfit.fit_files.modify_all(source=source, par=parname, column=col)
            
            # set bdfit p0 values
            if col != 'fixed':
                try:
                    self.dataline.bdfit.fitpar[col][parname] = \
                            float(self.parentry[parname][col][0].get())
                # failure cases: 
                #   KeyError on ncomp change
                #   ValueError on bad user input
                except (ValueError, KeyError):
                    pass
            
            elif col == 'fixed':
                try:
                    if self.parentry[parname][col][0].get():
                        self.bfit.fit_files.share_var[parname].set(False)
                except KeyError:
                    pass
        
        # set synchronization        
        for p in self.parentry.keys():
            parentry = self.parentry[p]
            
            # shared box value synchronization
            var = self.bfit.fit_files.share_var[p]
            parentry['shared'][0] = var
            parentry['shared'][1].config(variable=var)
                        
            # set callback
            for k in ('p0', 'blo', 'bhi', 'fixed'):
                        
                # remove old trace callbacks
                for t in parentry[k][0].trace_vinfo():
                    parentry[k][0].trace_vdelete(*t)
                
                # set new trace callback
                parentry[k][0].trace_id = parentry[k][0].trace("w", \
                                partial(callback, parname=p, col=k, source=self))
                parentry[k][0].trace_callback = \
                                partial(callback, parname=p, col=k, source=self)
                
        # disallow fixed shared parameters
        def callback2(*args, parname):
            parentry = self.parentry[parname]
            var = self.bfit.fit_files.share_var[parname]
            if var.get():
                parentry['fixed'][0].set(False)
                    
        for p in self.parentry.keys():
            parentry = self.parentry[p]
            share = parentry['shared'][0]
            share.trace_id = share.trace("w", partial(callback2, parname=p))
            share.trace_callback = partial(callback2, parname=p)
    
    # ======================================================================= #
    def get_new_parameters(self):
        """
            Fetch initial parameters from fitter, set to data.    
            
            plist: Dictionary of initial parameters {par_name:par_value}
        """
        
        run = self.dataline.id
        
        # get pointer to fit files object
        fit_files = self.bfit.fit_files
        fitter = fit_files.fitter
        ncomp = fit_files.n_component.get()
        fn_title = fit_files.fit_function_title.get()
        
        # get list of parameter names
        plist = list(fitter.gen_param_names(fn_title, ncomp))
        plist.sort()
        
        # check if we are using the fit results of the prior fit
        values = None
        res = self.bfit.data[run].fitpar['res']
        isfitted = any([res[k] for k in res]) # is this run fitted?
        
        if fit_files.set_prior_p0.get() and not isfitted:
            r = 0
            for rkey in self.bfit.data:
                data = self.bfit.data[rkey]
                res = data.fitpar['res']
                
                isfitted = any([res[k] for k in res]) # is the latest run fitted?
                if isfitted and data.run > r:
                    r = data.run
                    values = {k:(res[k], 
                                 data.fitpar['blo'][k], 
                                 data.fitpar['bhi'][k], 
                                 data.fitpar['fixed'][k]) for k in res}
                    parentry = self.bfit.fit_files.fit_lines[rkey].parentry
                    
        # get calcuated initial values
        if values is None:
            values = fitter.gen_init_par(fn_title, ncomp, self.bfit.data[run].bd, 
                                     self.bfit.get_asym_mode(fit_files))
            
        # set to data
        self.bfit.data[run].set_fitpar(values)
            
        return tuple(plist)
        
    # ======================================================================= #
    def grid(self, row):
        """Re-grid a dataline object so that it is in order by run number"""
        self.row = row
        self.fitframe.grid(column=0, row=row, sticky=(W, N))
        self.fitframe.update_idletasks()
           
    # ======================================================================= #
    def degrid(self):
        """Remove displayed dataline object from file selection. """
        
        self.logger.debug('Degridding fitline for run %s', self.dataline.id)
        self.fitframe.grid_forget()
        self.fitframe.update_idletasks()
    
    # ======================================================================= #
    def set_input(self, source_line, parameter, column, set_all):
        """
            Set the input value for a given parameter to match the value in 
            another fitline
            
            source_line: the fitline to copy
            parameter:   name of the parameter to copy
            column:      name of the column to copy
            set_all:     boolean corresponding to fit_files.set_as_group    
        """
        
        # get parameter entry line and sharing
        try:
            parentry = self.parentry[parameter]
            shared = parentry['shared'][0].get()
            source_entry = source_line.parentry[parameter]
        except KeyError:
            return
              
        # set value
        if set_all or shared:
            
            p = parentry[column][0]
            
            # remove the trace
            p.trace_vdelete("w", p.trace_id)
            
            # set the value
            p.set(source_entry[column][0].get())
    
            # add the trace back
            p.trace_id = p.trace("w", p.trace_callback)
    
    # ======================================================================= #
    def show_fit_result(self):
        
        self.logger.debug('Showing fit result for run %s', self.dataline.id)
        
        # Set up variables
        displays = self.parentry
        
        try:
            data = self.dataline.bdfit
        except KeyError:
            return
            
        try:
            chi = data.chi
        except AttributeError:
            return 
        
        # display
        for parname in displays.keys():
            disp = displays[parname]
            showstr = "%"+".%df" % self.bfit.rounding
            disp['res'][0].set(showstr % data.fitpar['res'][parname])
            disp['dres-'][0].set(showstr % data.fitpar['dres-'][parname])
            disp['dres+'][0].set(showstr % data.fitpar['dres+'][parname])
            
            if 'chi' in disp.keys():
                disp['chi'][0].set('%.2f' % chi)
                if float(chi) > self.bfit.fit_files.chi_threshold:
                    disp['chi'][1]['readonlybackground']='red'
                else:
                    disp['chi'][1]['readonlybackground']=colors.readonly
    
    # ======================================================================= #
    def show_fn_composition(self):
        """
            Draw window with function components and total
        """
        
        self.logger.info('Drawing fit composition for run %s', self.dataline.id)
        
        # get top objects
        fit_files = self.bfit.fit_files
        bfit = self.bfit
        
        # get fit object
        bdfit = self.dataline.bdfit
        
        # get base function 
        fn_name = fit_files.fit_function_title.get()
        
        # get number of components and parameter names
        ncomp = fit_files.n_component.get()
        pnames_single = fit_files.fitter.gen_param_names(fn_name, 1)
        pnames_combined = fit_files.fitter.gen_param_names(fn_name, ncomp)
        
        if '2' in bdfit.mode:
            fn_single = fit_files.fitter.get_fn(fn_name=fn_name, ncomp=1, 
                            pulse_len=bdfit.get_pulse_s(), 
                            lifetime=bd.life[bfit.probe_species.get()])
            fn_combined = fit_files.fitter.get_fn(fn_name=fn_name, ncomp=ncomp, 
                            pulse_len=bdfit.get_pulse_s(), 
                            lifetime=bd.life[bfit.probe_species.get()])
        else:
            fn_single = fit_files.fitter.get_fn(fn_name=fn_name, ncomp=1)
            fn_combined = fit_files.fitter.get_fn(fn_name=fn_name, ncomp=ncomp)
        
        # draw in redraw mode
        draw_mode = bfit.draw_style.get()
        bfit.draw_style.set('redraw')
        
        # draw the data
        omit = bdfit.omit.get()
        if omit == bfit.fetch_files.bin_remove_starter_line:
            omit = ''
            
        bfit.draw(bdfit, bfit.get_asym_mode(fit_files), rebin=bdfit.rebin.get(), 
                    option=omit, figstyle='fit', color='k')
        
        # get the fit results 
        results = {par:bdfit.fitpar['res'][par] for par in pnames_combined}
        
        # draw if ncomp is 1
        if ncomp == 1:
            bfit.draw_style.set('stack')
            fit_files.draw_fit(bdfit.id, 'fit', unique=False, label=fn_name)
            self.bfit.draw_style.set(draw_mode)
            return
        
        # draw baseline
        if 'baseline' in pnames_single:
            bfit.plt.axhline('fit', bdfit.id+'_base', results['baseline'], ls='--', zorder=6)
        
        # get x pts
        t, a, da = bdfit.asym(bfit.get_asym_mode(fit_files))
        fitx = np.linspace(min(t), max(t), fit_files.n_fitx_pts)
                
        # get x axis scaling
        if bdfit.mode in bfit.units: 
            unit = bfit.units[bdfit.mode]
            fitxx = fitx*unit[0]
        else:                   
            fitxx = fitx
        
        # draw the combined
        params = [results[name] for name in pnames_combined]
        
        bfit.plt.plot('fit', bdfit.id+'_comb', fitxx, fn_combined(fitx, *params), 
                            unique=False, label='Combined', zorder=5)
        
        # draw each component
        for i in range(ncomp):
            
            # get parameters
            params = [results[single+'_%d'%i] \
                        for single in pnames_single if single != 'baseline']
            
            if 'baseline' in pnames_single:
                params.append(results['baseline'])
            
            # draw     
            bfit.plt.plot('fit', bdfit.id+'_%d'%i, fitxx, fn_single(fitx, *params), 
                            unique=False, ls='--', label='%s %d'%(fn_name, i), zorder=6)
            
        # plot legend
        bfit.plt.legend('fit')
        
        # reset to old draw mode
        bfit.draw_style.set(draw_mode)



# Track which plots are active, which to draw in. 
# Derek Fujimoto
# June 2019

import matplotlib as mpl
import matplotlib.pyplot as plt

# =========================================================================== #
class PltTracker(object):
    """
        active:         dictionary, id number of active plot
        plots:          dictionary, list of plots drawn for type
    """
    
    # ======================================================================= #
    def __init__(self):
        
        # lists for tracking all plots 
        self.plots = {'inspect':[], 'data':[], 'fit':[], 'param':[], 'periodic':[]}
        
        # track the active plot 
        self.active = {'inspect':0, 'data':0, 'fit':0, 'param':0, 'periodic':0}
    
    # ======================================================================= #
    def _close_figure(self, event):
        """Remove figure from list"""
        
        # get number and style
        number = event.canvas.figure.number
        style = event.canvas.style
        
        # disconnect events
        event.canvas.mpl_disconnect(event.canvas.user_close)
        event.canvas.mpl_disconnect(event.canvas.user_active)
        
        # close the winow
        plt.figure(number).clf()
        plt.close(number)
        
        # remove from list 
        self.plots[style].remove(number)
        
        # reset active
        try:
            self.active[style] = self.plots[style][-1]
        except IndexError:
            self.active[style] = 0
                        
    # ======================================================================= #
    def _decorator(self, style, fn, *args, id=None, unique=True, **kwargs):
        """
            Function wrapper
            
            style: one of "data", "fit", "param"
            fn: matplotlib function to operate on 
            args: passed to fn
            kwargs: passed to fn
        """
        
        # switch 
        fig = plt.figure(self.active[style])
        ax = plt.gca()
            
        # clear old objects
        if unique and id is not None:
            self._remove_drawn_object(ax, id)
        
        # run function 
        output = fn(*args, **kwargs)
        
        # track the drawn object
        if id is not None:
            ax.draw_objs.setdefault(id, []).append(output) 
        
        return output
    
    # ======================================================================= #
    def _remove_drawn_object(self, ax, draw_id):
        """
            Remove an object labelled by draw_id from the figure, based on draw 
            style.
        """
        color=None
        
        # check if id is present in drawn data
        if draw_id in ax.draw_objs.keys():

            # get item
            for item in ax.draw_objs[draw_id]:
            
                # remove line
                try:
                    item[0].remove()
                except TypeError:
                    item.remove()
                else: 
                    # remove errorbars
                    if type(item) == mpl.container.ErrorbarContainer:    
                        for i in item[1]:   i.remove()
                        for i in item[2]:   i.remove()
                        del ax.containers[ax.containers.index(item)]
                        
                    # remove lines
                    else:
                        try:
                            del ax.lines[ax.lines.index(item)]
                        except ValueError:
                            pass
                del item 
            
            # clear labels
            del ax.draw_objs[draw_id]
        
    # ======================================================================= #
    def _set_hover_annotation(self, fig, ax):
        """
            Setup the annotate object for mouse hover drawing
        """
        if not hasattr(ax, 'hover_annot'):
            ax.hover_annot = ax.annotate('', 
                             xy=(0, 0), 
                             xytext=(-3, 20), 
                             textcoords='offset points', 
                             ha='right', 
                             va='bottom', 
                             bbox=dict(boxstyle='round, pad=0.1', 
                                       fc='white', 
                                       alpha=0.1), 
                             arrowprops=dict(arrowstyle = '->', 
                                             connectionstyle='arc3, rad=0'), 
                             fontsize='xx-small', 
                            )    
        ax.hover_annot.set_visible(False)
        
        fig.canvas.mpl_connect("motion_notify_event", \
                               lambda x: self._show_annot_on_hover(x, fig, ax))
        
    # ======================================================================= #
    def _show_annot_on_hover(self, event, fig, ax):
        """
            If the cursor is near enough to the line, show the annotation. 
            
            Citation: https://stackoverflow.com/a/47166787
        """
        vis = ax.hover_annot.get_visible()
        if event.inaxes == ax:
            for line in ax.lines:        
                cont, ind = line.contains(event)
                if cont:
                    self._update_hover_annot(ind, line, ax)
                    fig.canvas.draw_idle()
                    break
                else:
                    if vis:
                        ax.hover_annot.set_visible(False)
                        fig.canvas.draw_idle()
        
    # ======================================================================= #
    def _update_active_id(self, event):
        """
            Update the active figure id based on click event.
        """
        number = event.canvas.figure.number
        style = event.canvas.style
        self.active[style] = number
    
    # ======================================================================= #
    def _update_hover_annot(self, ind, line, ax):
        """
            Update the hover annotation with that object's label.
            
            Citation: https://stackoverflow.com/a/47166787
        """
    
        i = ind["ind"][0]
        x, y = line.get_data()
        annot = ax.hover_annot
        annot.xy = (x[i], y[i])
        
        if hasattr(line, 'annot_label'):
            text = line.annot_label[i]
        else:
            text = line.get_label()
            
        annot.set_text(text)
        annot.set_backgroundcolor(line.get_color())
        annot.get_bbox_patch().set_alpha(0.4)
        annot.set_visible(True)
    
    # ======================================================================= #
    def annotate(self, style, id, *args, unique=True, **kwargs):
        return self._decorator(style, plt.annotate, *args, id=id, unique=unique, **kwargs)
    
    # ======================================================================= #
    def axhline(self, style, id, *args, unique=True, **kwargs):
        return self._decorator(style, plt.axhline, *args, id=id, unique=unique, **kwargs)
        
    # ======================================================================= #
    def axvline(self, style, id, *args, unique=True, **kwargs):
        return self._decorator(style, plt.axvline, *args, id=id, unique=unique, **kwargs)
        
    # ======================================================================= #
    def clf(self, style):
        """Clear the figure for a given style"""
        out = self._decorator(style, plt.clf)
        ax = plt.gca()
        ax.draw_objs = {}
        
    # ======================================================================= # 
    def errorbar(self, style, id, x, y, yerr=None, xerr=None, fmt='', ecolor=None, 
                 elinewidth=None, capsize=None, barsabove=False, lolims=False, 
                 uplims=False, xlolims=False, xuplims=False, errorevery=1, 
                 capthick=None, *, data=None, unique=True, annot_label=None, 
                 **kwargs):
        """
            Plot data.
            
            style: one of "data", "fit", or "param"
            annot_label: list of annotations for each point
            other arguments: defaults for matplotlib.pyplot.plot
        """
        
        # get active for this style
        active_style = self.active[style]
        
        # make new figure if needed 
        if active_style == 0:   
            self.figure(style)
            active_style = self.active[style]
        fig = plt.figure(active_style)
        ax = fig.axes[0]
        
        # redraw old objects and lines
        if unique:  self._remove_drawn_object(ax, id)
        self._remove_drawn_object(ax, 'line')
        
        # set up labelling of line objects for mouseover
        if 'label' in kwargs:
            label = kwargs.pop('label')
        else:
            label = ''
        
        # draw in active style 
        obj = plt.errorbar(x, y, yerr=yerr, xerr=xerr, fmt=fmt, ecolor=ecolor, 
                     elinewidth=elinewidth, capsize=capsize, 
                     barsabove=barsabove, lolims=lolims, uplims=uplims, 
                     xlolims=xlolims, xuplims=xuplims, errorevery=errorevery, 
                     capthick=capthick, data=data, **kwargs)

        # label the line object
        ax.lines[-1].set_label(label)
        
        # set the annotation
        if annot_label is not None: 
            ax.lines[-1].annot_label = annot_label
            
        # save the drawn object to the file
        ax.draw_objs.setdefault(id, []).append(obj)
        
        return obj
    
    # ======================================================================= #
    def figure(self, style, **kwargs):
        """
            Make new figure.
            
            style: one of "data", "fit", or "param"
            kwargs: keyword arguments to pass to plt.figure
        """
        
        # make figure
        fig = plt.figure(**kwargs)
        
        # make events and save as canvas attribute
        fig.canvas.user_close = fig.canvas.mpl_connect('close_event', self._close_figure)
        fig.canvas.user_active = fig.canvas.mpl_connect('button_press_event', self._update_active_id)
        
        # set style
        fig.canvas.style = style
        
        # set window name 
        fig.canvas.set_window_title('Figure %d (%s)' % (fig.number, style.title()))
        
        # update lists
        self.plots[style].append(fig.number)
        self.active[style] = fig.number
        
        # track drawn objects
        ax = plt.gca()
        if not hasattr(ax, 'draw_objs'):
            ax.draw_objs = {}
        
        # set up the hover annotations 
        self._set_hover_annotation(fig, ax)
        
        return fig

    # ======================================================================= #
    def gca(self, style):
        if not self.plots[style]: self.figure(style)
        return self._decorator(style, plt.gca)
    
    # ======================================================================= #
    def gcf(self, style):
        if not self.plots[style]: self.figure(style)
        return self._decorator(style, plt.gcf)
    
    # ======================================================================= #
    def legend(self, style, *args, **kwargs):
        self._decorator(style, plt.legend, *args, **kwargs)
        
    # ======================================================================= #
    def plot(self, style, id, *args, scalex=True, scaley=True, data=None, unique=True, **kwargs):
        """
            Plot data.
            
            style: one of "data", "fit", or "param"
            other arguments: defaults for matplotlib.pyplot.plot
        """
        
        # get active for this style
        active_style = self.active[style]
        
        # make new figure if needed 
        if active_style not in self.plots[style]:   
            self.figure(style)
        
        # draw in active style 
        fig = plt.figure(active_style)
        ax = fig.axes[0]
        
        # redraw old objects and lines
        if unique:  self._remove_drawn_object(ax, id)
        self._remove_drawn_object(ax, 'line')
        
        obj = plt.plot(*args, scalex=scalex, scaley=scaley, data=data, **kwargs)
        ax.draw_objs.setdefault(id, []).append(obj)
        
        return obj

    # ======================================================================= #
    def text(self, style, *args, id=None, unique=True, **kwargs):
        return self._decorator(style, plt.text, *args, id=id, unique=unique, **kwargs)

    # ======================================================================= #
    def tight_layout(self, style, *args, **kwargs):
        return self._decorator(style, plt.tight_layout, *args, **kwargs)

    # ======================================================================= #
    def xlabel(self, style, *args, **kwargs):
        return self._decorator(style, plt.xlabel, *args, **kwargs)
    
    # ======================================================================= #
    def xlim(self, style, *args, **kwargs):
        return self._decorator(style, plt.xlim, *args, **kwargs)
    
    # ======================================================================= #
    def xticks(self, style, *args, **kwargs):
        return self._decorator(style, plt.xticks, *args, **kwargs)
    
    # ======================================================================= #
    def ylabel(self, style, *args, **kwargs):
        return self._decorator(style, plt.ylabel, *args, **kwargs)
    
    # ======================================================================= #
    def ylim(self, style, *args, **kwargs):
        return self._decorator(style, plt.ylim, *args, **kwargs)
    
    # ======================================================================= #
    def yticks(self, style, *args, **kwargs):
        return self._decorator(style, plt.yticks, *args, **kwargs)
    
    # ======================================================================= #
    def zlabel(self, style, *args, **kwargs):
        return self._decorator(style, plt.zlabel, *args, **kwargs)
    
    # ======================================================================= #
    def zlim(self, style, *args, **kwargs):
        return self._decorator(style, plt.zlim, *args, **kwargs)
    
    # ======================================================================= #
    def zticks(self, style, *args, **kwargs):
        return self._decorator(style, plt.zticks, *args, **kwargs)
    



# Calculate needed current from desired magnetic field. 
# Derek Fujimoto
# December 2017

from tkinter import *
from tkinter import ttk
import numpy as np
import webbrowser
import logging
import bfit
from bfit import logger_name

# =========================================================================== #
class calculator_nqr_B0(object):
    
    # ======================================================================= #
    def __init__(self):
        """Draw window for Zaher's calculator"""
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # root 
        root = Toplevel()
        root.title("Zaher's Caculator")
        root.columnconfigure(0, weight=1)
        root.rowconfigure(0, weight=1)

        # icon
        try:
            img = PhotoImage(file=bfit.icon_path)
            root.tk.call('wm', 'iconphoto', root._w, img)
        except Exception as err:
            print(err)

        # key bindings
        root.bind('<Return>', self.calculate)             
        root.bind('<KP_Enter>', self.calculate)
        
        # variables
        self.field = StringVar()
        self.field.set("")
        self.current = StringVar()
        self.current.set("")
        
        # main frame
        mainframe = ttk.Frame(root, pad=5)
        mainframe.grid(column=0, row=0, sticky=(N, W))
        mainframe.columnconfigure(0, weight=1)
        mainframe.rowconfigure(0, weight=1)
        
        # Entry and other objects
        title_line = ttk.Label(mainframe,   
                text='BNQR Magnetic Static Field -- Current Converter', 
                justify=CENTER)
        self.entry_field = ttk.Entry(mainframe, textvariable=self.field, width=10, 
                justify=RIGHT)
        gauss = ttk.Label(mainframe, text='Gauss')
        equals = ttk.Label(mainframe, text='=')
        self.entry_current = ttk.Entry(mainframe, 
                textvariable=self.current, width=10, justify=RIGHT)
        amperes = ttk.Label(mainframe, text='Amperes')
        explanation = ttk.Label(mainframe, 
                text='Press Enter to convert', 
                justify=CENTER)
                
        link = ttk.Button(mainframe, 
                          text='Calibration Data Here', 
                          command=lambda : webbrowser.open(\
                                  'http://bnmr.triumf.ca/?file=HH_Calibration'))
        
        # Gridding
        title_line.grid(        column=0, row=0, padx=5, pady=5, columnspan=5)
        self.entry_field.grid(  column=0, row=1, padx=5, pady=5)
        gauss.grid(             column=1, row=1, padx=5, pady=5)
        equals.grid(            column=2, row=1, padx=20, pady=5)
        self.entry_current.grid(column=3, row=1, padx=5, pady=5)
        amperes.grid(           column=4, row=1, padx=5, pady=5)
        explanation.grid(       column=0, row=2, padx=5, pady=5, columnspan=5)
        link.grid(              column=0, row=3, padx=5, pady=5, columnspan=5)
        
        # runloop
        self.root = root
        self.logger.debug('Initialization success. Starting mainloop.')
        root.mainloop()
        
    # ======================================================================= #
    def calculate(self, *args):
        
        # check focus
        focus_id = str(self.root.focus_get())
        
        # convert field to current
        if focus_id == str(self.entry_field):        
            try:
                field = float(self.field.get()) 
                value = field2current(field)
                self.current.set("%.4f" % np.around(value, 4))
                self.logger.info('Field of %g G converted to current of %g A', 
                                 field, value)
            except ValueError:
                self.logger.exception('Bad input')
                self.current.set('Error')
            
        # convert current to field
        elif focus_id == str(self.entry_current):        
            try:
                current = float(self.current.get()) 
                value = current2field(current)
                self.field.set("%.4f" % np.around(value, 4))
                self.logger.info('Current of %g A converted to field of %g G', 
                                 current, value)
            except ValueError:
                self.logger.exception('Bad input')
                self.field.set('Error')
            
# ======================================================================= #
def current2field(current):    return current*2.2131+0.175
def field2current(field):      return (field-0.175)/2.2131




# Drawing style window
# Derek Fujimoto
# July 2018

from tkinter import *
from tkinter import ttk
from bfit import logger_name
import logging
import webbrowser

# ========================================================================== #
class popup_drawstyle(object):
    """
        Popup window for setting drawing options. 
    """

    # ====================================================================== #
    def __init__(self, parent):
        self.parent = parent
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
         # make a new window
        self.win = Toplevel(parent.mainframe)
        self.win.title('Set Marker and Line Styles')
        frame = ttk.Frame(self.win, relief='sunken', pad=5)
        
        # icon
        parent.set_icon(self.win)
        
        # Key bindings
        self.win.bind('<Return>', self.set)             
        self.win.bind('<KP_Enter>', self.set)
        
        # initialize 
        self.entry_list = []
        self.label_list = []
        self.text_list = []
        self.key_list = list(parent.style.keys())
        self.key_list.sort()
        
        # add entries
        for i, key in enumerate(self.key_list):
            
            # create objects
            self.label_list.append(ttk.Label(frame, text=key, pad=5, justify=LEFT))
            self.text_list.append(StringVar())
            self.entry_list.append(ttk.Entry(frame, 
                    textvariable=self.text_list[-1], width=20, justify=RIGHT))
            self.entry_list[-1].insert(0, parent.style[key])
            
            # grid objects
            self.label_list[-1].grid(column=0, row=i, sticky=W)
            self.entry_list[-1].grid(column=1, row=i, columnspan=2)
          
        # add buttons
        set_button = ttk.Button(frame, text='Set', command=self.set)
        close_button = ttk.Button(frame, text='Cancel', command=self.cancel)
        help_button = ttk.Button(frame, text='Help', command=self.help)
        set_button.grid(column=0, row=i+1)
        help_button.grid(column=1, row=i+1)
        close_button.grid(column=2, row=i+1)
            
        # grid frame
        frame.grid(column=0, row=0)
        self.logger.debug('Initialization success. Starting mainloop.')
        
    # ====================================================================== #
    def help(self):
        self.logger.info('Opening help webpage.')
        webbrowser.open('https://matplotlib.org/api/_as_gen/matplotlib.lines.'+\
                        'Line2D.html#matplotlib.lines.Line2D')

    # ====================================================================== #
    def set(self, *args):
        """Set entered values"""
        for k, t in zip(self.key_list, self.text_list):
            
            val = t.get()
            
            if type(self.parent.style[k]) == float:
                val = float(val)
            
            self.parent.style[k] = val
        self.logger.info('Setting values %s to %s', self.key_list, 
                                            [s.get() for s in self.text_list])
        self.win.destroy()
        
    # ====================================================================== #
    def cancel(self):
        self.win.destroy()



# Calculate needed current from desired magnetic field. 
# Derek Fujimoto
# December 2017

from tkinter import *
from tkinter import ttk
import numpy as np
import pandas as pd
import logging
import bfit
import os
from bfit import logger_name
from bfit.backend import colors

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure

# =========================================================================== #
class calculator_nmr_atten(object):
    
    data_location = os.path.join(os.path.dirname(__file__), 
                                '..',
                                'data', 
                                'nmr_antenna_data.csv')
    
    # ======================================================================= #
    def __init__(self):
        """Draw window for Iain's calculator"""
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # root 
        root = Toplevel()
        root.title("Iain's Caculator")
        root.columnconfigure(0, weight=1)
        root.rowconfigure(0, weight=1)

        # icon
        try:
            img = PhotoImage(file=bfit.icon_path)
            root.tk.call('wm', 'iconphoto', root._w, img)
        except Exception as err:
            print(err)

        # key bindings
        root.bind('<Return>', self.calculate)             
        root.bind('<KP_Enter>', self.calculate)
        
        # variables
        self.power = StringVar()
        self.power.set("")
        self.dac = StringVar()
        self.dac.set("")
        
        # load data
        self.data = pd.read_csv(self.data_location, comment="#")
        
        # normalize 
        df_max = self.data['antenna_amplitude (mV)'].max()
        self.data['power (%)'] = self.data['antenna_amplitude (mV)']/df_max*100
        
        # main frame
        mainframe = ttk.Frame(root, pad=5)
        mainframe.grid(column=0, row=0, sticky=(N, W))
        mainframe.columnconfigure(0, weight=1)
        mainframe.rowconfigure(0, weight=1)
        
        # Entry and other objects
        title_line = ttk.Label(mainframe,   
                text='B-NMR RF Power -- DAC Linear Interpolation', 
                justify=CENTER)
        self.entry_power = ttk.Entry(mainframe, textvariable=self.power, width=10, 
                justify=RIGHT)
        percent = ttk.Label(mainframe, text='RF Power (%)')
        equals = ttk.Label(mainframe, text='=')
        self.entry_dac = ttk.Entry(mainframe, 
                textvariable=self.dac, width=10, justify=RIGHT)
        dac = ttk.Label(mainframe, text='DAC Setpoint')
        explanation = ttk.Label(mainframe, 
                text='Press Enter to convert', 
                justify=CENTER)
        
        # Gridding
        c = 0
        title_line.grid(        column=0, row=0, padx=5, pady=5, columnspan=5)
        percent.grid(           column=c, row=1, padx=5, pady=5); c+=1
        self.entry_power.grid(  column=c, row=1, padx=5, pady=5); c+=1
        equals.grid(            column=c, row=1, padx=20, pady=5); c+=1
        self.entry_dac.grid(    column=c, row=1, padx=5, pady=5); c+=1
        dac.grid(               column=c, row=1, padx=5, pady=5); c+=1
        explanation.grid(       column=0, row=2, padx=5, pady=5, columnspan=5)
        
        # embed the data figure in the tk window
        fig = Figure(figsize=(6, 4), dpi=100)
        ax = fig.add_subplot(111)
        
        ax.plot(self.data['rf_level_control (DAC)'], self.data['power (%)'], 'o', 
                color=colors.foreground)
        ax.set_ylabel('Normalized Antenna Readback (%)', fontsize='small', 
                color=colors.foreground)
        ax.set_xlabel('rf_level_control (DAC Setpoint)', fontsize='small', 
                color=colors.foreground)
        
        # text
        ax.text(0, 5, 'Data taken Oct 2018, M1354', color=colors.foreground, 
                fontsize='small')
        
        # set plot elements
        ax.set_facecolor(colors.background)
        fig.set_facecolor(colors.background)
        for loc in ['bottom', 'top', 'right', 'left']:
            ax.spines[loc].set_color(colors.foreground)
        ax.tick_params(axis='both', labelsize='small', colors=colors.foreground)
        fig.tight_layout()
        
        canvas = FigureCanvasTkAgg(fig, master=root)  # A tk.DrawingArea.
        canvas.draw()
        
        canvas.get_tk_widget().grid(column=0, row=3, padx=5, pady=5, columnspan=5)
        
        # runloop
        self.root = root
        self.logger.debug('Initialization success. Starting mainloop.')
        root.mainloop()
        
    # ======================================================================= #
    def calculate(self, *args):
        
        # check focus
        focus_id = str(self.root.focus_get())
        
        # convert power to dac
        if focus_id == str(self.entry_power):        
            try:
                power = float(self.power.get()) 
                value = self.power2dac(power)
                self.dac.set("%d" % value)
                self.logger.info('Power of %g converted to dac setpoint of %d', 
                                 power, value)
            except ValueError:
                self.logger.exception('Bad input')
                self.dac.set('Error')
            
        # convert dac to power
        elif focus_id == str(self.entry_dac):        
            try:
                dac = int(self.dac.get()) 
                value = self.dac2power(dac)
                self.power.set("%.4f" % np.around(value, 4))
                self.dac.set("%d" % int(dac))
                self.logger.info('dac setpoint of of %d converted to power of %g', 
                                 dac, value)
            except ValueError:
                self.logger.exception('Bad input')
                self.power.set('Error')
            
    # ======================================================================= #
    def power2dac(self, value): 
        if value == 0:
            return 2047
        else:
            self.data.sort_values('power (%)', inplace=True)
            return int(np.interp(value, 
                             self.data['power (%)'].values, 
                             self.data['rf_level_control (DAC)'].values,
                             left=0,
                             right=2047)
                        )

    def dac2power(self, value): 
        self.data.sort_values('rf_level_control (DAC)', inplace=True)
        return np.interp(value, 
                         self.data['rf_level_control (DAC)'],
                         self.data['power (%)'],
                         left=100,
                         right=0)
                         





import setuptools
from distutils.core import Extension
from Cython.Build import cythonize
import numpy, os

with open("README.md", "r") as fh:
    long_description = fh.read()

# module extension
ext = Extension("bfit.fitting.integrator",
                sources=["./bfit/fitting/integrator.pyx",
                        "./bfit/fitting/FastNumericalIntegration_src/integration_fns.cpp"],
                language="c++",             # generate C++ code                        
                include_dirs=["./bfit/fitting/FastNumericalIntegration_src",numpy.get_include()],
                libraries=["m"],
                extra_compile_args=["-ffast-math"]
                )

setuptools.setup(
    name="bfit",
    version="4.3.2",
    author="Derek Fujimoto",
    author_email="fujimoto@phas.ubc.ca",
    description="BNMR/BNQR Data Fitting and Visualization",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dfujim/bfit",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Cython",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Development Status :: 5 - Production/Stable",
    ],
    install_requires=['cython>=0.28','numpy>=1.14','tqdm>=4.25.0',
                      'bdata>=6.1.4','matplotlib>=2.2.4','pandas>=0.23.0',
                      'pyyaml>=5.1','scipy>=1.2.0','iminuit>=1.5.2'],
    package_data={'': ['./data']},
    include_package_data=True,
    ext_modules = cythonize([ext], include_path=[numpy.get_include()]),
)



# Fit set of run combined asymmetry globally 
# Derek Fujimoto
# Nov 2018

from bfit.fitting.global_fitter import global_fitter
import bdata as bd
import numpy as np
import collections

# =========================================================================== #
class global_bdata_fitter(global_fitter):
    
    # ======================================================================= #
    def __init__(self, data, fn, xlims=None, rebin=1, asym_mode='c', **kwargs):
        """
            data:       list of bdata objects
            
            fn:         list of function handles to fit (or single which applies to all)
                        must specify inputs explicitly (do not do def fn(*par)!)
                        must have len(fn) = len(runs) if list
                        
            xlims:      list of 2-tuples for (low, high) bounds on fitting range 
                            based on x values. If list is not depth 2, use this 
                            range on all runs.
            
            rebin:      rebinning factor on fitting and drawing data
            
            fixed:      list of booleans indicating if the paramter is to be 
                        fixed to p0 value (same length as p0). Returns best 
                        parameters in order presented, with the fixed 
                        parameters omitted.
            
            asym_mode:  asymmetry type to calculate and fit
            
            kwargs:     passed to global_fitter
        """
        
        # check input type
        if type(data) in (bd.bdata, bd.bjoined, bd.bmerged):
            data = [data]
        ndata = len(data)
        
        # Set rebin
        if not isinstance(rebin, collections.Iterable):
            rebin = [rebin]*ndata
        
        # Get asymmetry
        asym = [d.asym(asym_mode, rebin=re) for d, re in zip(data, rebin)]
        
        # split into x, y, dy data sets
        x = [a[0] for a in asym]
        y = [a[1] for a in asym]
        dy = [a[2] for a in asym]
        
        # select subrange
        if xlims is not None:
            
            # check depth
            if len(np.array(xlims).shape) < 2:
                xlims = [xlims]*ndata
                            
            # initialize new inputs
            xnew = []
            ynew = []
            dynew = []
            
            # select subrange
            for i, xl in enumerate(xlims):
                tag = (xl[0]<x[i])*(x[i]<xl[1])
                xnew.append(x[i][tag])
                ynew.append(y[i][tag])
                dynew.append(dy[i][tag])
            
            # new arrays
            x = xnew
            y = ynew
            dy = dynew
            
        # intialize
        super(global_bdata_fitter, self).__init__(fn, x, y, dy=dy, **kwargs)
